/*
 TinySVM -- Yet Another Tiny SVM Package

 $Id: classifier.cpp,v 1.4 2002/08/20 06:31:16 taku-ku Exp $;

 Copyright (C) 2001-2002  Taku Kudo <taku-ku@is.aist-nara.ac.jp>
 All rights reserved.

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Library General Public
 License as published by the Free Software Foundation; either
 version 2 of the License, or (at your option) any later verjsion.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Library General Public License for more details.

 You should have received a copy of the GNU Library General Public
 License along with this library; if not, write to the
 Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 Boston, MA 02111-1307, USA.
*/


#include "common.h"
#include "classifier.h"
#include "freqtm.h"

namespace TinySVM {

Classifier::Classifier (const BaseExample & example, const Param & param):
  Kernel (example, param)
{
  try {
     x = example.x;
     y = example.y;
     model_bias = 0;

    if (feature_type == BINARY_FEATURE && dot_kernel) {
      _getDistance = &Classifier::_getDistance_binary;

      dot_buf = new int[l];
      binary_kernel_cache = new double *[l];
      fi2si = new int *[d + 1];
      int *_fisize = new int[d + 1];

      // init
      for (int i = 0; i < (d + 1); i++)
	_fisize[i] = 0;

      // 1st, detemin the size of each array
      for (int i = 0; i < l; i++) {
	feature_node *node = (feature_node *) x[i];
	int feature_num = 0;
	while (node->index >= 0) {
	  _fisize[node->index]++;
	  node++;
	  feature_num++;
	}

	// caluculate cache
	model_bias += y[i] * (this->*_getKernel2) (0.0);
	binary_kernel_cache[i] = new double[feature_num + 1];
	for (int j = 1; j <= feature_num; j++) {
	  binary_kernel_cache[i][j] =
	    y[i] * ((this->*_getKernel2) ((double) j) -
		    (this->*_getKernel2) ((double) (j - 1)));
	}
      }

      // 2nd memory allocate for each array
      for (int i = 0; i < (d + 1); i++) {
	fi2si[i] = new int[_fisize[i] + 1];
	_fisize[i] = 0;
      }

      // 3rd, copy value
      for (int i = 0; i < l; i++) {
	feature_node *node = (feature_node *) x[i];
	while (node->index >= 0) {
	  fi2si[node->index][_fisize[node->index]++] = i;
	  node++;
	}
      }

      // final fill -1 (dumy field)
      for (int i = 0; i < (d + 1); i++)
	fi2si[i][_fisize[i]++] = -1;

      delete[]_fisize;

    } else if ( (kernel_type == TREE 
                 || kernel_type == TREE_FAST
                 || kernel_type == TREE_COLLINS 
                 || kernel_type == TREE_COLLINS_FAST
                 || kernel_type == TREE_COLLINS_P
                 || kernel_type == TREE_COLLINS_P_FAST )
                && node_restriction_type == RESTRICT_1 &&
                geometric_type == NO_GEOMETRIC ) {
      _getDistance = &Classifier::_getDistance_tree_restrict1;
      /*
      freqtm 
        = new FreqtmTreeRestrict1Classification((BaseExample&)example, 
                                                2, // minsup
                                                1, // minpat
                                                0xffffffff, // maxpat
                                                20, // malicious_size
                                                depth_factor, 
                                                false // verbose
                                                );
      */
      // _getDistance = &Classifier::_getDistance_normal;
    } else {
      _getDistance = &Classifier::_getDistance_normal;
    }
  }

  catch (...) {
    fprintf (stderr, "Classifier::Classifier(): Out of memory\n");
    exit (EXIT_FAILURE);
  }
}

Classifier::~Classifier ()
{
  if (feature_type == BINARY_FEATURE && dot_kernel) {
    delete [] dot_buf;
    for (int i = 0; i < l; i++)
      delete [] binary_kernel_cache[i];
    delete [] binary_kernel_cache;

    for (int i = 0; i < (d + 1); i++)
      delete [] fi2si[i];
    delete [] fi2si;
  }
}

double
Classifier::_getDistance_normal (const feature_node * _x) const
{
  register double result = 0.0;
  register feature_node *node = fix_feature_node ((feature_node *) _x);

  for (register int i = 0; i < l; i++)
    //result += y[i] * (this->*_getKernel) (x[i], node);
    result += y[i] * getKernel(node, x[i]);
  
  return result;
}

double
Classifier::_getDistance_tree_restrict1 (const feature_node * _x) const
{
  register double result    = 0.0;
  // register feature_node *cx = fix_feature_node ((feature_node *) _x);

  if ( Param::is_node_lookup_kernel( kernel_type ) ) {
    if ( _x->extra->getTree() ) {
      _x->extra->getTree()->initNodeLookup();
    }
    if ( _x->extra->getTreeCollinsP() ) {
      _x->extra->getTreeCollinsP()->initNodeLookup();
    }
  }
  
  if ( _x->extra->getTreeVec() ) {
    // in training
    double* dest = new double[ l ];
    getKernel( _x, x, l, dest ); // possiblly efficient
    double kernel_sum = 0.0;
    for ( int k = 0; k < l ; k++ ) {
      kernel_sum += y[k] * dest[k];
    }
    
    if ( false ) {
      feature_node* n = _x->extra->getTreeVec();
      _x->extra->setTreeVec( 0 );
      double kernel_sum_direct = 0.0;
      for ( int k = 0; k < l ; k++ ) {
        double dk = getKernel( _x, x[k] );
        kernel_sum_direct += y[k] * dk;
        double diff = (dk-dest[k]);
        if ( k <= 10 || diff != 0.0 ) {
          double esim = _x->extra->getTree()->getEdgePatternSimilarity( x[k]->extra->getTree() );
          std::cerr << "diff=" << (dk-dest[k]) << " k=" << dk << " k(v)=" << dest[k] << " esim=" << esim << "\n" << std::flush; 
        }
      }
      std::cerr << "sum=" << kernel_sum_direct << " sum(v)=" << kernel_sum << "\n" << std::flush;
      _x->extra->setTreeVec( n );
    }

    delete[] dest;
    return kernel_sum;
  } else {
    return _getDistance_normal( _x );
   }
}


double
Classifier::_getDistance_binary (const feature_node * node) const
{
  register double result = model_bias;
  register int j, k;
  register int *p;

  memset (dot_buf, 0, sizeof (int) * l);

  feature_node *org = (feature_node *) node;
  while (node->index >= 0 && node->index <= d) {
    if (node->value != 1)
      return this->_getDistance_normal (org);
    for (p = fi2si[node->index], j = 0; (k = p[j]) != -1; j++)
      result += (binary_kernel_cache[k][++dot_buf[k]]);
    node++;
  }

  return result;
}

};
