/*
 This file is added by J. Kazama to the TinySVM package.
 Copyright (C) 2005. Jun'ichi Kazama <kazama@jaist.ac.jp>
 All rights reserved.

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Library General Public
 License as published by the Free Software Foundation; either
 version 2 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Library General Public License for more details.

 You should have received a copy of the GNU Library General Public
 License along with this library; if not, write to the
 Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 Boston, MA 02111-1307, USA.
*/

#ifndef _FREQTM_H
#define _FREQTM_H


/*
 The FREQTM algorithm, a modified FREQT.
 FREQTM enumerates all frequent subtrees in a given set of trees, and 
 converts them into feature vectors using the subtrees as the features,
 while avoiding exponential enumeration.

 The aim is to make the training with tree kernels substantially faster.
 
 FREQTM is based on the FREQT algorithm (Asai et al., 2002). 
 Howerver, the difference is as follows.  
 - FREQTM detects malicious tree pairs which cause exponential enumeration. 
   If two trees are very similar, original FREQT easily cause
   exponential enumeration of subtrees. 
   FREQTM detects such malicious tree pairs and never enumerate subtrees 
   that occur only in that trees. 

 The kernel between malicious pairs is calclulated using the standard DP 
 procedures, while the kernel values for other pairs can be calculated 
 using the inner product on the converted vectores.
 (i.e., faster training is possible)

 To understand the speed-up method, you will also need to read kernel.h

 As for the driver code (i.e., the "main" code), see tree2fv.cpp

 For more details, please refer to our paper:
   "Speeding up Training with Tree Kernels for Node Relation Labeling"
   Jun'ichi Kazama and Kentaro Torisawa, HLT-EMNLP 2005.
 
 The code is heavily based on Dr. Taku Kudo's FREQT implementation.
 (I thank him very much for his outstanding research and softwares!)
 However, of course, I added and restructured fair amount of codes for
 my purposes. So, some of the bugs will be attributed not to the original 
 implementation but to my implementation.

 I cleaned-up the code after HLT-EMNLP 05 experiments, so this is not exactly
 the same as the code used in the experiments.
*/

#include <iostream>
#include <vector>
#include <string>
#include <map>
#include <set>
#include <ext/hash_map>
#include <ext/hash_set>
#include <iterator>
#include <cstdlib>
#include <cstring>
#include <unistd.h>

#include "../config.h"
#include "base_example.h"
#include "lotree.h"
#include "gp.h" 

using namespace std;
using namespace TinySVM;


template< bool isMarked, typename T >
class NodePrinter
{
};

template< typename T >
class NodePrinter< false, T >
{
 public:
  static void print( ostream& os, const T& n ) {
    os << '(';
#ifdef LABEL_INTEGER
    os << '[' << n << ']';
#else
    os << '[' << LOTreeNode::escape_char( n ) << ']';
#endif
  }
};

template< typename T >
class NodePrinter< true, T >
{
 public:
  static void print ( ostream& os, const pair< T, int >& n ) {
    if ( n.second == -1 ) {
      os << '(';
    } else {
      os << '(' << '*' << n.second;
    }
#ifdef LABEL_INTEGER
    os << '[' << n.first << ']';
#else
    os << '[' << LOTreeNode::escape_char( n.first ) << ']';
#endif
  }
};


/** 
 *  Base class of FREQTM
 */
class FreqtmBase
{
 public:
  int     verbose;
  typedef unsigned long biguint;
#ifdef LABEL_INTEGER
  typedef int label_t;
  static const label_t return_label = -1;
#else
  typedef string label_t;
  label_t return_label;
#endif

  BaseExample&             examples;
  vector < const LOTree* > transaction;
  
  vector< vector< pair<unsigned int, double> > > feature_vector;
  
  typedef
  vector< set< unsigned int > > malicious_pair_t;
  malicious_pair_t              malicious_pair;
  
  // occurence pattern (effective featre) => (regist_no, feature weight)
  typedef 
  vector< pair<unsigned int, unsigned int> > occ_pattern_t;
  
  typedef
  map< occ_pattern_t, pair< unsigned int, double> > occ_map_t;
  occ_map_t effective_occ;
  vector< pair< occ_pattern_t, double> > occ_vec;

  mutable vector< unsigned int >  tmp_uniq_loc_v;
  mutable occ_pattern_t           tmp_occ_v;
  
  ostream* feature_weight_stream;
  ostream* pattern_out_stream;
  
  // control parameters
  unsigned int minsup;
  unsigned int minpat;
  unsigned int maxpat;
  unsigned int malicious_size;
  double       malicious_rate;
  double       malicious_rate_inv;
  bool         do_precheck;
  bool         do_heavy_check;
  bool         do_feature_compress;
  bool         do_track_pattern;
  bool         do_output_pattern;
  double       depth_factor;
  
  // some states
  bool              in_precheck;
  vector< biguint > feature_vector_size;
  vector< biguint > feature_vector_comp_size;
  int               max_pattern_size;
  int               max_malicious_size;
  int               max_malicious_pattern_size;  
  biguint           enumeration_no;
  biguint           feature_no;
  biguint           feature_no_comp;
  bool              once_run;
  
  
  void setPatternOutStream( ostream* os ) {
    pattern_out_stream = os;
  }
  
  void setFeatureWeightStream( ostream* os ) {
    feature_weight_stream = os;
  }

  FreqtmBase( BaseExample& _examples,
              unsigned int _minsup,
              unsigned int _minpat,
              unsigned int _maxpat,
              unsigned int _malicious_size,
              double       _malicious_rate,
              bool _do_precheck,
              bool _do_heavy_check,
              bool _do_feature_compress,
              bool _do_track_pattern,
              bool _do_output_pattern,
              double _depth_factor, int _verbose )
    : examples( _examples ),
      transaction( _examples.l ),
      malicious_pair( _examples.l ),
      pattern_out_stream( 0 ),
      feature_weight_stream( 0 ),
      feature_vector_size( _examples.l ),
      feature_vector_comp_size( _examples.l ),
      feature_vector( _examples.l )
  {
    
#ifndef LABEL_INTEGER
    return_label = ")";
#endif
    verbose  = _verbose;
    minsup   = _minsup;
    minpat   = _minpat;
    maxpat   = _maxpat;
    malicious_size = _malicious_size;
    malicious_rate = _malicious_rate;
    malicious_rate_inv = 1.0 / (double)( _malicious_rate );
    do_precheck         = _do_precheck;
    do_heavy_check      = _do_heavy_check;
    do_feature_compress = _do_feature_compress;
    do_track_pattern    = _do_track_pattern;
    do_output_pattern   = _do_output_pattern;
    depth_factor        = _depth_factor;

    if ( do_output_pattern && !do_track_pattern ) {
      cerr << "Enable do_track_pattern when using do_output_pattern\n" << flush;
      exit(1);
    }
    
    if ( verbose > 0 ) {
      cerr << "sizeof(biguint) = " << sizeof( biguint ) << "\n" << flush;
    }
    
    for ( int i = 0; i < examples.l; i++ ) {
      if ( examples.x[i]->extra 
           && examples.x[i]->extra->getTree() ) {
        transaction[i] = examples.x[i]->extra->getTree();
        //cerr << transaction[i]->toString() << "\n" << flush;
      } else {
        cerr << "ERROR: there is no tree part.\n" << flush;
        exit(1);
      }
    }
    
    if ( verbose > 0 ) {
      cerr << "# of transaction = " << transaction.size() << "\n" << flush;
    }
    
    once_run = false;
  }
  
  virtual void run() = 0;
};



/**
 * Base Implementation class of FREQTM
 */
template< bool isMarked, bool doMarkPruning >
class FreqtmImplTypes
{
 public:
  class projected_t_marked
  {
  public:
    bool marked;
    projected_t_marked() : marked( false ) { }
  };
  
  class projected_t_markpruning
  {
  public:
    bool marked;
    vector< unsigned int > roots;
    projected_t_markpruning() : marked( false ), roots( 0 ) { }
  };
  
  class projected_t 
    : public gp::Select<isMarked, 
                        typename gp::Select<doMarkPruning, 
                                            projected_t_markpruning,
                                            projected_t_marked>::Result,
                        gp::EmptyType>::Result
  {
  public:
    typedef
    typename gp::Select<isMarked, 
                        typename gp::Select<doMarkPruning, 
                                            projected_t_markpruning,
                                            projected_t_marked>::Result,
                        gp::EmptyType>::Result Parent;
    
    int                    depth;
    unsigned int           support;
    unsigned int           node_num;
    double                 weight;
    bool                   initialized;
    vector< unsigned int > locations;
    vector< unsigned int > positions;
    
    projected_t() : Parent(), 
                    depth( 0 ), support( 0 ), node_num( 0 ), 
                    weight( 1.0 ), initialized( false ) 
    {       
    }
  };
  
  typedef  FreqtmBase::label_t label_t;
  typedef
  typename gp::Select< isMarked, 
                       pair< label_t, int >, 
                       label_t >::Result node_t;
  
  // type for storing expansion points (depends on marked or not)
  // pair of ( how many we track back from the first expansion point, 
  //           expanded_label(plus mark) )
  typedef pair< int, node_t > expand_t;
  typedef vector < expand_t > pattern_t;
};

template< bool isMarked, bool doMarkPruning, typename CandidateKeyType >
class FreqtmImpl : public FreqtmBase
{
 public:
  typedef FreqtmImplTypes<isMarked, doMarkPruning> Types;
  typedef typename Types::projected_t projected_t;
  typedef typename Types::node_t      node_t;
  typedef typename Types::expand_t    expand_t;
  typedef typename Types::pattern_t   pattern_t;
  typedef CandidateKeyType candidate_key_t;
  typedef map< candidate_key_t, projected_t > candidate_t;

  pattern_t pattern;
  //mutable pattern_t   tmp_prefix; 
  vector< pattern_t > effective_pattern;
 
  pattern_t           max_pattern;
  pattern_t           max_malicious_pattern;
  
  FreqtmImpl( BaseExample& _examples,
              unsigned int _minsup,
              unsigned int _minpat,
              unsigned int _maxpat,
              unsigned int _malicious_size,
              double       _malicious_rate,
              bool _do_precheck,
              bool _do_heavy_check,
              bool _do_feature_compress,
              bool _do_track_pattern,
              bool _do_output_pattern,
              double _depth_factor, int _verbose )
    : FreqtmBase( _examples, _minsup, _minpat, _maxpat, 
                  _malicious_size, _malicious_rate, 
                  _do_precheck, _do_heavy_check, _do_feature_compress,
                  _do_track_pattern, _do_output_pattern, _depth_factor, 
                  _verbose ) 
  {
    if ( verbose > 0 ) {
      cerr << "size of projected_t =" 
           << sizeof( projected_t ) << "\n" << flush;
    }
  }

  virtual ~FreqtmImpl() 
  {

  }
  
  void print_pattern_to_stream( ostream& os,  
                                pattern_t& pattern) const 
  {
    int n = 0;
    for (unsigned int i = 0; i < pattern.size(); ++i) {
      if ( pattern[i].first != 0 ) {
        for ( int k = 0; k < pattern[i].first; k++ ) {
          os << ')';
          --n;
        }
      } else {
        ++n;
        NodePrinter< isMarked, label_t >::print( os, pattern[i].second );
      }
    }
    for (int i = 0; i < n; ++i) os << ')';
  }
  
  
  inline
  unsigned int wsupport( const projected_t &projected ) const
  {
    return projected.locations.size();
  }
  
  inline
  unsigned int support( const projected_t &projected ) const 
  {
    unsigned int old = 0xffffffff; // INT_MAX
    unsigned int size = 0; 
    for (unsigned int i = 0; i < projected.locations.size(); ++i) {
      if (projected.locations[i] != old) ++size;
      old = projected.locations[i];
    }
    return size;
  }
  
  inline
  const vector< unsigned int >&
  get_uniq_locations( const projected_t& projected,
                      vector< unsigned int >& to ) const
  {
    if ( projected.support == wsupport( projected ) ) {
      return projected.locations;
    }
    to.resize( 0 );
    unsigned int old = 0xffffffff; // INT_MAX
    unsigned int size = 0; 
    for (unsigned int i = 0; i < projected.locations.size(); ++i) {
      if (projected.locations[i]  != old) {
        ++size;
        to.push_back( projected.locations[i] );
      }
      old = projected.locations[i];
    }
    return to;
  }
  
  inline
  const vector< pair<unsigned int, unsigned int> >&
  get_occ_pattern( const projected_t& projected,
                   vector< pair<unsigned int, unsigned int> >& to ) const
  {
    to.resize( 0 );
    unsigned int old = 0xffffffff; // INT_MAX
    unsigned int size = 0; 
    for (unsigned int i = 0; i < projected.locations.size(); ++i) {
      if (projected.locations[i] != old) {
        ++size;
        to.push_back( pair<unsigned int, unsigned int>(projected.locations[i], 1) );
      }
      else {
        to[ size - 1 ].second++;
      }
      old = projected.locations[i];
    }
    return to;
  }
  
  void register_projected ( const projected_t& projected ) 
  {
    if ( projected.node_num < minpat ) return;
    
    //print_pattern_to_stream( cerr, pattern ); cerr << "\n" << flush;
    if ( (int)projected.node_num > max_pattern_size ) {
      max_pattern_size = projected.node_num;
      max_pattern      = pattern;
    }
    
    const vector< unsigned int >& locs
      = get_uniq_locations( projected, tmp_uniq_loc_v );
    
    const vector< pair< unsigned int, unsigned int > >& occ 
      = get_occ_pattern( projected, tmp_occ_v );
    
    unsigned int pre_feature_no_comp = feature_no_comp;
    
    for (unsigned int i = 0; i < locs.size(); i++) {
      feature_vector_size[ locs[i] ]++;
    }
    if ( do_feature_compress ) {
      occ_map_t::iterator oi = effective_occ.find( occ );
      if ( oi == effective_occ.end() ) {
        pair< unsigned int, double >& ent = effective_occ[occ];
        ent.first  = feature_no_comp;
        ent.second = projected.weight;
        if ( do_output_pattern ) {
          effective_pattern.push_back( pattern );
        }
        for (unsigned int i = 0; i < locs.size(); i++) {
          feature_vector_comp_size[ locs[i] ]++;
        }
        feature_no_comp++; 
      } else {
        oi->second.second += projected.weight;
        if ( do_output_pattern ) {
          // register most
          if (pattern.size() < effective_pattern[ oi->second.first ].size()) {
            effective_pattern[ oi->second.first ] = pattern;
          }
        }
      }
      feature_no++;
    } else {
      occ_vec.push_back( make_pair( occ, projected.weight ) );
      for ( unsigned int i = 0; i < occ.size(); i++ ) {
        pair<unsigned int, double> 
          feat( feature_no, occ[i].second ); 
        feature_vector[ occ[i].first ].push_back( feat ); 
      }
      if ( do_output_pattern ) {
        if ( pattern_out_stream ) {
          *pattern_out_stream 
            << feature_no << " " << projected.weight << "\n";
          print_pattern_to_stream( *pattern_out_stream, pattern );
          *pattern_out_stream << "\n";
          
        }
      }
      feature_no++;
    }
    
    if ( verbose > 0 ) {
      if ( do_feature_compress ) {
        if ( pre_feature_no_comp != feature_no_comp ) {
          if ( feature_no_comp % 100000 == 0 ) {
            cerr << "[" << feature_no_comp << "]" << flush;
          } else if ( feature_no_comp % 10000 == 0 ) {
            cerr << "*" << flush;
          }
        }
      } else {
        if ( feature_no % 100000 == 0 ) {
          cerr << "[" << feature_no << "]" << flush;
        } else if ( feature_no % 10000 == 0 ) {
          cerr << "*" << flush;
        }
      }
    }
  }
  
  inline void prune( candidate_t& candidate ) const
  {
    for( typename candidate_t::iterator l = candidate.begin(); 
         l != candidate.end(); ) {
      unsigned int wsup = wsupport( l->second );
      if ( wsup < minsup ) {
        candidate.erase( l++ );
      } else {
        unsigned int sup= support( l->second );
        if (sup < minsup) {
          candidate.erase( l++ );
        } else {
          l->second.support = sup;
          ++l;
        }
      }
    }
    return;
  }
 
  inline
  void register_malicious( const projected_t &projected )
  {
    if ( do_precheck && !in_precheck ) {
      // no need to register (since already registered in precheck)
      return;
    }
    if ( (int)projected.support > max_malicious_size ) {
      max_malicious_size    = projected.support;
      max_malicious_pattern = pattern;
    }
    const vector< unsigned int >& locs 
      = get_uniq_locations( projected, tmp_uniq_loc_v );
    
    for ( unsigned int i = 0; i < locs.size(); i++ ) {
      for ( unsigned int j = locs.size()-1; j >= i+1; j-- ) {
        malicious_pair[ locs[i] ].insert( locs[j] );
      }
    }
  }

  inline
  bool is_malicious( const projected_t &projected ) const 
  {
    const vector< unsigned int >& locs 
      = get_uniq_locations( projected, tmp_uniq_loc_v );
    for ( unsigned int i = 0; i < locs.size(); i++ ) {
      for ( unsigned int j = locs.size()-1; j >= i+1; j-- ) {
        if ( malicious_pair[ locs[i] ].count( locs[j] ) == 0 ) {
          return false;
        }
      }
    } 
    return true;
  }
  
  inline double duplicate_rate( const projected_t& projected ) const 
  {
    return ( (double)projected.support/(double)projected.locations.size() );
  }

  // Behaviour switching functions
  //
  inline void register_proj( const projected_t& projected,
                             gp::Int2Type< false > ) {
    register_projected( projected );
  }

  inline void register_proj( const projected_t& projected,
                             gp::Int2Type< true > ) {
    if ( projected.marked ) { register_projected( projected ); }
  }

 
  inline void extract_node( node_t& node, const LOTreeNode* l,
                            gp::Int2Type< false > ) const {
    // not marked
    node = l->getLabel();
  }
  
  inline void extract_node( node_t& node, const LOTreeNode* l,
                            gp::Int2Type< true > ) const {
    node.first  = l->getLabel();
    node.second = l->getMark();
  }
  
  inline void update_mark( projected_t& tmp, 
                           const projected_t& projected,
                           const LOTreeNode* l,
                           gp::Int2Type< false > ) const {
    // do nothing
  }
  
  inline void update_mark( projected_t& tmp, 
                           const projected_t& projected,
                           const LOTreeNode* l,
                           gp::Int2Type< true > ) const {
    tmp.marked  = projected.marked || (l->isMarked());
  }

  inline void update_mark( projected_t& tmp, 
                           const projected_t& projected,
                           bool mark,
                           gp::Int2Type< false > ) const {
    // do nothing
  }
  
  inline void update_mark( projected_t& tmp, 
                           const projected_t& projected,
                           bool mark,
                           gp::Int2Type< true > ) const {
    tmp.marked  = projected.marked || mark;
  }


  inline bool check_markpruning( const projected_t& projected,
                                 unsigned int root_no, 
                                 const LOTreeNode* l,
                                 gp::Int2Type< false > ) const {
    return false;
  }

  inline bool check_markpruning( const projected_t& projected,
                                 unsigned int root_no, 
                                 const LOTreeNode* l,
                                 gp::Int2Type< true > ) const {
    if ( ! projected.marked &&
         ! l->isMarked() ) {
      const LOTreeNode* nn = l->getNearestMarkedNode();
      if ( nn == 0 || nn->getNoPostOrdered() > root_no ) {
        return true;
      }
    }
    return false;
  }

  inline unsigned int get_root_no( const projected_t& projected, 
                                   unsigned int i,
                                   gp::Int2Type< false > ) const {
    return 0;
  }

  inline unsigned int get_root_no( const projected_t& projected,
                                   unsigned int i,
                                   gp::Int2Type< true > ) const {
    return projected.roots[i];
  }
  
  inline void push_back_root( projected_t& tmp, unsigned int rn,
                              gp::Int2Type< false > ) const {
    // do nothing
  }

  inline void  push_back_root( projected_t& tmp, unsigned int rn,
                              gp::Int2Type< true > ) const {
    tmp.roots.push_back( rn );
  }
  //
  //
  

  inline virtual 
  void push_back_pattern( pattern_t& pattern, const candidate_key_t& key ) = 0;

  bool project( const candidate_key_t& key,
                const projected_t& projected )
  {
    enumeration_no++;
    if ( verbose > 0 ) {
      if ( enumeration_no % 100000 == 0 ) {
        cerr << "(" << enumeration_no << ")" << flush;
      } else if ( enumeration_no % 10000 == 0 ) {
        cerr << "." << flush;
      }
    }

    unsigned int previous_pattern_size;

    if ( do_track_pattern ) {
      previous_pattern_size = pattern.size();
      push_back_pattern( pattern, key );
    }

    unsigned int current_sup = projected.support;

    if ( projected.node_num >= malicious_size ) {      
      register_malicious( projected );
      return false;
    }
    if (projected.node_num >= maxpat ) {
      if ( !in_precheck ) {
        register_proj( projected, gp::Int2Type<isMarked>() );
      } 
      return true;
    }
    
    candidate_t candidate;

    if ( !generate_candidate( projected, candidate ) ) {
      register_malicious( projected );
      return false;
    }
   
    prune ( candidate );

    for ( typename candidate_t::const_iterator l = candidate.begin (); 
          l != candidate.end(); ++l) {
      if (do_heavy_check && is_malicious( l->second ) ) {
        continue;
      }
      
      bool suc = project( l->first, l->second );
      
      if ( !suc && current_sup == l->second.support ) {
        return false;
      }
    }

    if ( !in_precheck ) {
      register_proj( projected, gp::Int2Type<isMarked>() );
    }

    if ( do_track_pattern ) {
      pattern.resize( previous_pattern_size );
    }

    return true;
  }  

  virtual void run() 
  {
    if ( once_run ) {
      cerr << "You can't run more than once.\n";
      exit(1);
    }

    enumeration_no     = 0;
    feature_no         = 0;
    feature_no_comp    = 0;

    max_malicious_size         = 0;
    max_pattern_size           = -1;
    max_malicious_pattern_size = -1;
    in_precheck = false;
    // collect size1 patterns
    candidate_t size1;
    generate_size1_candidate( size1 );    
    cerr << "size1.size() = " << size1.size() << "\n" << flush;
    prune ( size1 );
    cerr << "size1.size() [after pruning] = " << size1.size() << "\n" << flush;
    if ( do_precheck ) {
      in_precheck = true;
      if ( verbose > 0 ) { cerr << "start precheck\n" << flush; }
      
      for (typename candidate_t::const_iterator l = size1.begin (); 
           l != size1.end(); ++l) {
        project ( l->first, l->second ); 
      }
      cerr << "\n" << flush;
      in_precheck = false;
      enumeration_no = 0;
    }
    
    if ( verbose > 0 ) { cerr << "start enumeration\n" << flush; }
    for (typename candidate_t::const_iterator l = size1.begin (); 
         l != size1.end(); ++l) {
      project ( l->first, l->second );
    }
    if ( verbose > 0 ) {  cerr << "\n" << flush; }

    post_run();
    
    once_run = true;
    return;
  }

  virtual void post_run() {
    if ( do_feature_compress ) {
      // make feature vector from occurence map
      for ( unsigned int i = 0; i < feature_vector.size(); i++ ) {
        feature_vector[i].resize( feature_vector_comp_size[i] );
        feature_vector[i].resize( 0 );
      }
      unsigned int fno = 0;
      for ( occ_map_t::const_iterator itr = effective_occ.begin();
            itr != effective_occ.end(); itr++ ) {
        const vector< pair< unsigned int, unsigned int > >& 
          occ = itr->first;
        for ( unsigned int i = 0; i < occ.size(); i++ ) {
          pair< unsigned int, double>
            feat( fno, occ[i].second );
          feature_vector[ occ[i].first ].push_back( feat );
        }
        // write pattern
        if ( do_output_pattern && pattern_out_stream ) {
          *pattern_out_stream << fno << " " << itr->second.second << "\n";
          print_pattern_to_stream( *pattern_out_stream,
                                   effective_pattern[ itr->second.first ] );
          *pattern_out_stream << "\n";
        }
        fno++;
      }
      if ( fno != feature_no_comp ) {
        cerr << "Invalid # of feature (compress)\n" << flush;
        exit(1);
      }
      // write feature weight
      if ( feature_weight_stream ) {
        *feature_weight_stream << feature_no_comp << "\n";
        fno = 0;
        for ( occ_map_t::const_iterator itr = effective_occ.begin();
            itr != effective_occ.end(); itr++ ) {
          const occ_pattern_t& occ_pat = itr->first;
          *feature_weight_stream            
            << fno << " " << itr->second.second << " " << occ_pat.size();
          for( int k = 0; k < occ_pat.size(); k++ ) {
            *feature_weight_stream 
              << " " << occ_pat[k].first << " " << occ_pat[k].second; 
          }
          *feature_weight_stream << "\n";
          fno++;
        }
      } else {
        cerr << "Can't write feature weights.\n" << flush;
      }
    } else {
      if ( feature_weight_stream ) {
        *feature_weight_stream << feature_no << "\n";
        for ( int i = 0; i < occ_vec.size(); i++ ) {
          const occ_pattern_t& occ_pat = occ_vec[i].first;
          double weight = occ_vec[i].second;
          *feature_weight_stream            
            << i << " " << weight << " " << occ_pat.size();
          for( int k = 0; k < occ_pat.size(); k++ ) {
            *feature_weight_stream 
              << " " << occ_pat[k].first << " " << occ_pat[k].second; 
          }
          *feature_weight_stream << "\n";
        }
      } else {
        cerr << "Can't write feature weights.\n" << flush;
      }
    }

    // check feature vector
    for ( unsigned int i = 0; i < feature_vector.size(); i++ ) {
      bool ok = true;
      if ( do_feature_compress ) {
        if ( feature_vector[i].size() != feature_vector_comp_size[i] ) {
          ok = false;
        }
      } else {
        if ( feature_vector[i].size() != feature_vector_size[i] ) {
          ok = false;
        }
      }
      if ( !ok ) {
        cerr << "Incorrect feature vectors.\n" << flush;
        exit(1);
      }
    }

    // 
    for ( int i = 0; i < examples.l; i++ ) {
      feature_node* example = examples.x[i];

      // set absolute exaple id
      example->extra->setNo( i );


      vector< pair<unsigned int, double> >& v = feature_vector[i];
      feature_node* tree_vec = new feature_node[ v.size() + 1 ];
      for( unsigned int j = 0; j < v.size(); j++ ) {
        tree_vec[j].index = v[j].first;
        tree_vec[j].value = v[j].second;
      }
      tree_vec[ v.size() ].index = -1;
      tree_vec[ v.size() ].value = 0;
      example->extra->setTreeVec( tree_vec );

      if ( malicious_size != 0xffffffff ) {
        // malicious_size is set       
        example->extra->setMaliciousPair( i );
        set< unsigned int >& mal_pair = malicious_pair[i];
        for ( set< unsigned int >::const_iterator itr = mal_pair.begin(); 
              itr != mal_pair.end(); itr++ ) {
          example->extra->setMaliciousPair( *itr );
          examples.x[ *itr ]->extra->setMaliciousPair( i );
        }
      }
    }

    // report statistics
    int max_size 
      = *std::max_element( feature_vector_size.begin(), 
                           feature_vector_size.end());
    int min_size 
      = *std::min_element( feature_vector_size.begin(), 
                           feature_vector_size.end());
    int max_comp_size 
      = *std::max_element( feature_vector_comp_size.begin(), 
                           feature_vector_comp_size.end());
    int min_comp_size 
      = *std::min_element( feature_vector_comp_size.begin(), 
                           feature_vector_comp_size.end());
    
    biguint sum = 0;
    for (unsigned int i = 0; i < feature_vector_size.size(); i++) {
      sum += feature_vector_size[i];
    }
    biguint sum_comp = 0;
    for (unsigned int i = 0; i < feature_vector_comp_size.size(); i++) {
      sum_comp += feature_vector_comp_size[i];
    }
    biguint node_sum = 0;
    double  node_sum2 = 0.0;
    for ( unsigned int i = 0; i < transaction.size(); i++ ) {
      node_sum += transaction[i]->numNodes();
      node_sum2 += (transaction[i]->numNodes() * transaction[i]->numNodes());
    }
    double ave_tree_size  = node_sum  / (double)transaction.size();
    double ave_tree_size2 = node_sum2 / (double)transaction.size();
    cerr << "# trees              = " << transaction.size() << "\n";
    cerr << "avg tree size        = " << ave_tree_size << "\n";
    cerr << "avg (tree size)^2    = " << ave_tree_size2 << "\n";
    cerr << "# features           = " << feature_no << "\n"; 
    cerr << "max |feature vector| = " << max_size << "\n";
    cerr << "min |feature vector| = " << min_size << "\n";
    cerr << "avg |feature vector| = " << (double)sum/(double)(feature_vector_size.size()) << "\n";
    cerr << "# features          (compressed)  = " 
         << feature_no_comp << "\n";           
    cerr << "max |feature vector (compressed)| = " 
         << max_comp_size << "\n";
    cerr << "min |feature vector (compressed)| = " 
         << min_comp_size << "\n";
    cerr << "avg |feature vector (compressed)| = " 
         << (double)sum_comp/(double)(feature_vector_comp_size.size()) << "\n";

    cerr << "Max pattern size  = " << max_pattern_size << "\n" << flush;
    cerr << "      the pattern = ";
    print_pattern_to_stream( cerr, max_pattern );
    cerr << "\n" << flush;
    cerr << "Max malicious set = " << max_malicious_size << "\n" << std::flush;
    cerr << "      the pattern = ";
    print_pattern_to_stream( cerr, max_malicious_pattern );
    cerr << "\n" << flush;
    
    biguint malp_num = 0;
    for( unsigned int i = 0; i < transaction.size(); i++) {
      malp_num += malicious_pair[i].size();
    }
    malp_num *= 2;
    malp_num += transaction.size();

    double mal_rate = (double)malp_num/((double)transaction.size()*(double)transaction.size());
    cerr << "# malicious pair = " << malp_num << "\n" << flush;
    cerr << "            rate = " << mal_rate << "\n" << flush;
  }

  virtual bool 
  generate_candidate( const projected_t& projected,
                      candidate_t& candidate ) const = 0;
  virtual void 
  generate_size1_candidate( candidate_t& size1 ) const = 0;
  
};


// Corresponds to Kashima's tree kernel
template< bool isMarked, bool doMarkPruning >
class FreqtmTypes
{
 public:
  typedef 
  typename FreqtmImplTypes<isMarked,doMarkPruning>::expand_t
  candidate_key_t;
};

template< bool isMarked, bool doMarkPruning >
class Freqtm 
  : public FreqtmImpl< isMarked, doMarkPruning,
                       typename FreqtmTypes<isMarked,doMarkPruning>::candidate_key_t >
{
 public:
  
  typedef FreqtmImpl< isMarked, doMarkPruning,
                      typename FreqtmTypes<isMarked,doMarkPruning>::candidate_key_t > Parent;
  
  // template does not inherit typedefs ...
  typedef typename Parent::projected_t     projected_t; // projected_t;
  typedef typename Parent::expand_t        expand_t;
  typedef typename Parent::pattern_t       pattern_t;
  typedef typename Parent::candidate_key_t candidate_key_t;
  typedef typename Parent::candidate_t     candidate_t;
  
  // Since GCC 3.4, 
  // template does not show base-template-class's members ...
  using Parent::transaction;
  using Parent::malicious_rate_inv;
  using Parent::depth_factor;

  Freqtm( BaseExample& _examples,
          unsigned int _minsup,
          unsigned int _minpat,
          unsigned int _maxpat,
          unsigned int _malicious_size,
          double       _malicious_rate,
          bool         _do_precheck,
          bool         _do_heavy_check,
          bool         _do_feature_compress,
          bool         _do_track_pattern,
          bool         _do_output_pattern,
          double _depth_factor, int _verbose)
    : Parent( _examples, _minsup, _minpat, _maxpat, 
              _malicious_size, _malicious_rate,
              _do_precheck, _do_heavy_check, _do_feature_compress, 
              _do_track_pattern, _do_output_pattern, 
              _depth_factor, _verbose )
  {
    
  }
  
  virtual ~Freqtm() { }
  
  virtual bool 
  generate_candidate( const projected_t& projected,
                      candidate_t& candidate ) const
  {
    int depth = projected.depth;
    //pattern_t& prefix = tmp_prefix;
    expand_t prefix;
    
    double support_inv = 1.0 / (double)projected.support;

    for (unsigned int i = 0; i < projected.locations.size(); ++i) {
      unsigned int  id     = projected.locations[i];
      int          pos     = projected.positions[i];
      unsigned int root_no = get_root_no( projected, i, 
                                          gp::Int2Type< doMarkPruning >() );
      const LOTree* tree = transaction[id];
      const LOTreeNode* pnode  = tree->getNodePostOrdered(pos);
      for (int d = -1 ; d < depth && pnode != 0; ++d) {
	const LOTreeNode* start 
          = (d == -1) ? pnode->getFirstChild() : pnode->getNextSibling();
	int newdepth = depth - d;
        prefix.first = d+1;
	for ( const LOTreeNode* l = start; l != 0; l = l->getNextSibling() ) {
          
          extract_node( prefix.second, l, gp::Int2Type< isMarked >() );
          if ( check_markpruning( projected, root_no, l, 
                                  gp::Int2Type< doMarkPruning >() ) ) {
            continue;
          }
          
	  projected_t &tmp = candidate[ prefix ];
	  tmp.locations.push_back( id );
          tmp.positions.push_back( l->getNoPostOrdered() );
          
          // for mark pruning 
          push_back_root( tmp, root_no, gp::Int2Type< doMarkPruning >() );

          if ( tmp.locations.size() * support_inv >  malicious_rate_inv ) {
            return false;
          }
          
          if ( !tmp.initialized ) {
            tmp.depth        = newdepth;
            tmp.node_num     = projected.node_num + 1;
            update_mark( tmp, projected, l, gp::Int2Type< isMarked >() );
            tmp.weight       = projected.weight * depth_factor;
            tmp.initialized  = true;
          }
        }
	if (d != -1) pnode = pnode->getParent();
      }
    }
    return true;
  }

  virtual void 
  generate_size1_candidate( candidate_t& size1 ) const
  {
    for (unsigned int i = 0; i < transaction.size(); ++i) {
      const LOTree* tree = transaction[i];
      for (int j = 0; j < tree->numNodes(); ++j) {
        const LOTreeNode* node = tree->getNodePostOrdered( j );
        expand_t ep;
        ep.first = 0;
        extract_node( ep.second, node, gp::Int2Type< isMarked >() );
        projected_t& tmp = size1[ ep ];
	tmp.locations.push_back ( i );
        tmp.positions.push_back ( j );
        
        // for mark pruning (if required)
        push_back_root( tmp, j, gp::Int2Type< doMarkPruning >() );

        if ( !tmp.initialized ) {
          tmp.depth        = 0;
          tmp.node_num     = 1;
          update_mark( tmp, tmp, node, gp::Int2Type< isMarked >() );
          tmp.weight       = depth_factor;
          tmp.initialized  = true;
        }
      }
    }
  }

  inline 
  virtual void push_back_pattern( pattern_t& pattern,
                                  const candidate_key_t& key ) {
    pattern.push_back( key );
  }
};



/**
 * For the Collins' kernel 
 */
template< bool isMarked, bool doMarkPruning>
class FreqtmCollinsTypes
{
 public:
  typedef FreqtmImplTypes<isMarked, doMarkPruning> ParentTypes;
  typedef
  vector< typename ParentTypes::node_t >   children_pattern_t;
  typedef
  vector< typename ParentTypes::expand_t > expands_t;
  typedef 
  pair<expands_t, children_pattern_t> candidate_key_t;
};

template< bool isMarked, bool doMarkPruning >
class FreqtmCollins 
  : public FreqtmImpl< isMarked, doMarkPruning,
                       typename FreqtmCollinsTypes<isMarked,doMarkPruning>::candidate_key_t >
{
 public:
  typedef
  FreqtmImpl< isMarked, doMarkPruning,
              typename FreqtmCollinsTypes<isMarked,doMarkPruning>::candidate_key_t > Parent;
  typedef FreqtmCollinsTypes< isMarked, doMarkPruning > Types;

  typedef typename Types::children_pattern_t children_pattern_t;
  typedef typename Types::expands_t    expands_t;
  typedef typename Parent::node_t      node_t;
  typedef typename Parent::pattern_t   pattern_t;
  typedef typename Parent::projected_t projected_t; // projected_t;
  typedef typename Parent::expand_t    expand_t;
  typedef typename Parent::candidate_key_t candidate_key_t;
  typedef typename Parent::candidate_t candidate_t;

  // Since GCC 3.4, 
  // template does not show base-template-class's members ...
  using Parent::transaction;
  using Parent::malicious_rate_inv;
  using Parent::depth_factor;
  using Parent::minpat;
  using Parent::max_malicious_size;
  using Parent::pattern;

  FreqtmCollins( BaseExample& _examples,
                 unsigned int _minsup,
                 unsigned int _minpat,
                 unsigned int _maxpat,
                 unsigned int _malicious_size,
                 double       _malicious_rate,
                 bool _do_precheck,
                 bool _do_heavy_check,
                 bool _do_feature_compress,
                 bool _do_track_pattern,
                 bool _do_output_pattern,
                 double _depth_factor, int _verbose)
    : Parent( _examples, _minsup, _minpat, _maxpat, 
              _malicious_size, _malicious_rate,
              _do_precheck, _do_heavy_check, 
              _do_feature_compress, 
              _do_track_pattern, _do_output_pattern,
              _depth_factor, _verbose )
  {
    //
    if ( minpat < 2 ) { 
      minpat = 2; 
      cerr << "minpat is changed to 2 for Collins kernels" << endl << flush; 
    }
  }

  virtual ~FreqtmCollins() { }
  
  void complement_pattern( const projected_t& projected,
                           pattern_t&         pattern ) const 
  {
    unsigned int id  = projected.locations[0]; // use one of tree
    int         pos  = projected.positions[0];
    const LOTree*     tree  = transaction[id];
    const LOTreeNode* pnode = tree->getNodePostOrdered(pos);
    int depth = projected.depth;
    expand_t ep;
    int      k = 0;
    for (int d = 0 ; d < depth && pnode != 0; ++d) {
      for ( const LOTreeNode* l = pnode->getNextSibling(); 
            l != 0; l = l->getNextSibling() ) {
        ep.first = k + 1;
        extract_node( ep, l, gp::Int2Type< isMarked >() );
        k = 0;
      }
      k++;
      pnode = pnode->getParent();
    }
  }

  inline void 
  register_projected ( const projected_t& projected ) 
  {
    // this test is redundant but may be OK.
    if ( projected.node_num < minpat ) return;

    unsigned int pre_pattern_size = pattern.size();
    complement_pattern( projected, pattern );
    Parent::register_projected( projected );
    pattern.resize( pre_pattern_size );
  } 
  
  inline void 
  register_malicious( const projected_t &projected )
  {
    if ( (int)projected.support > max_malicious_size ) {
      unsigned int pre_pattern_size = pattern.size();
      complement_pattern( projected, pattern );
      Parent::register_malicious( projected );
      pattern.resize( pre_pattern_size );
    } else {
      Parent::register_malicious( projected );
    }
  }

  inline 
  bool generate_candidate_sub( unsigned int id,
                               const projected_t&  projected,
                               unsigned int        root_no, 
                               const LOTreeNode*   parent,
                               int                 new_depth,
                               expands_t&          prefix,
                               children_pattern_t& suffix,
                               candidate_t& candidate ) const
  {
    double support_inv = 1.0 / (double)projected.support;

    const LOTreeNode* start = parent->getFirstChild(); 
    if ( start ) { // there is the first child (production)
      expand_t ep;
      ep.first = 0;
      extract_node( ep.second, start, gp::Int2Type<isMarked>() ); 
      prefix.push_back( ep );
      suffix.resize( 0 );
      node_t node;

      // I expect this is removed by optimization if not used
      bool marked = start->isMarked();
      for ( const LOTreeNode* l = start->getNextSibling(); 
            l != 0; l = l->getNextSibling() ) {
        extract_node( node, l, gp::Int2Type<isMarked>() );
        suffix.push_back( node );
        marked = (marked || l->isMarked());
      }
      if ( check_markpruning( projected, root_no, start, gp::Int2Type< doMarkPruning>() ) ) {
        prefix.pop_back();
        return true;
      }
      projected_t& tmp = candidate[ make_pair( prefix, suffix ) ];
      tmp.locations.push_back( id );
      tmp.positions.push_back( start->getNoPostOrdered() );

      // for mark pruning 
          push_back_root( tmp, root_no, gp::Int2Type< doMarkPruning >() );

      if ( tmp.locations.size() * support_inv >  malicious_rate_inv ) {
        return false;
      }

      if ( !tmp.initialized ) {
        tmp.depth    = new_depth;
        tmp.node_num = projected.node_num + 1 + suffix.size();
        update_mark( tmp, projected, marked, gp::Int2Type<isMarked>() );
        tmp.weight   = projected.weight * depth_factor;
        tmp.initialized = true;
      }
      prefix.pop_back();
    }
    return true;
  }

  bool generate_candidate( const projected_t &projected,
                           candidate_t& candidate ) const
  {
    int         depth  = projected.depth;
    expands_t   prefix; 
    children_pattern_t  suffix;
    expand_t ep;

    for (unsigned int i = 0; i < projected.locations.size(); ++i) {
      unsigned int  id     = projected.locations[i];
      int          pos     = projected.positions[i];
      unsigned int root_no = get_root_no( projected, i, 
                                          gp::Int2Type< doMarkPruning >() );
      const LOTree*     tree  = transaction[id];
      const LOTreeNode* pnode = tree->getNodePostOrdered(pos);
      
      prefix.resize( 0 );
      if (!generate_candidate_sub( id, projected, root_no, pnode, 
                                   depth+1, prefix, suffix, 
                                   candidate ) ) {
        return false;
      }
      int k = 0;
      
      for ( int d = 0; d < depth && pnode != 0; ++d ) {
        for( const LOTreeNode* p = pnode->getNextSibling();
             p != 0; p = p->getNextSibling() ) {
          ep.first = k+1;
          extract_node( ep.second, p, gp::Int2Type<isMarked>() );
          prefix.push_back( ep );

          if (!generate_candidate_sub( id, projected, root_no, p, 
                                       depth-d+1, prefix, suffix,
                                       candidate )) {
            return false;
          }
          k = 0;
        }
        pnode = pnode->getParent();
        k++;
      }
    }
    return true;
  }
 
  void 
  generate_size1_candidate( candidate_t& size1 ) const
  {
    for (unsigned int i = 0; i < transaction.size(); ++i) {
      const LOTree* tree = transaction[i];
      for (int j = 0; j < tree->numNodes(); ++j) {
        const LOTreeNode* node = tree->getNodePostOrdered( j );

        expands_t pv(1); 
        pv[0].first = 0; 
        extract_node( pv[0].second, node, gp::Int2Type<isMarked>() );
        children_pattern_t sv(0);
        projected_t& tmp = size1[ make_pair(pv,sv) ];
	tmp.locations.push_back ( i );
        tmp.positions.push_back ( j );
        
        // for mark pruning (if required)
        push_back_root( tmp, j, gp::Int2Type< doMarkPruning >() );

        if ( !tmp.initialized ) {
          tmp.depth        = 0;
          tmp.node_num     = 1;
          update_mark( tmp, tmp, node, gp::Int2Type<isMarked>() );
          tmp.weight       = 1.0;
          tmp.initialized  = true;
        }
      }
    }
  }

  inline 
  virtual void push_back_pattern( pattern_t& pattern,
                                  const candidate_key_t& key ) 
  {
    // key is pair< vector<expand_t>, children_pattern_t>
    for ( typename vector<expand_t>::const_iterator 
            l = key.first.begin(); l != key.first.end(); l++ ) {
      pattern.push_back( *l );
    }
  }
};

#endif
