/*
 TinySVM -- Yet Another Tiny SVM Package

 $Id: kernel.h,v 1.18 2002/08/20 06:31:17 taku-ku Exp $;

 Copyright (C) 2001-2002  Taku Kudo <taku-ku@is.aist-nara.ac.jp>
 All rights reserved.

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Library General Public
 License as published by the Free Software Foundation; either
 version 2 of the License, or (at your option) any later verjsion.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Library General Public License for more details.

 You should have received a copy of the GNU Library General Public
 License along with this library; if not, write to the
 Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 Boston, MA 02111-1307, USA.
*/


#ifndef _KERNEL_H
#define _KERNEL_H
#include "misc.h"
#include "param.h"
#include "base_example.h"
#include "lotree.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

using namespace std;

// $Id: kernel.h,v 1.18 2002/08/20 06:31:17 taku-ku Exp $;
namespace TinySVM {

class Kernel
{
protected:
  // constant param
  const int    l;
  const int    d;
  const int    pack_d;
  const int    kernel_type;
  const int    feature_type;
  const int    dot_kernel;
  const int    degree;
  const double param_g;
  const double param_r;
  double param_s;

  // J.K.
  mutable int          max_tree_size;
  mutable double**     tree_c;
  mutable double**     tree_s;
  mutable double**     tree_ca;
  mutable double**     tree_cr;
  mutable double**     tree_sr;
  mutable double**     tree_car;

  const int    verbose;
  const bool   tree_check; // check the correctness of tree kernels
  const int    geometric_type;
  const int    node_restriction_type;
  const double restriction_ratio;
  const bool   self_normalize;
  //const bool   nural_normalize;
  const double depth_factor;
  const int    threshold_normalize;
  const double threshold_normalize_threshold;
  const bool   ignore_malicious;
  const bool   ignore_trie;

  const BaseExample&           example;
  const vector< double >& feature_weight;
  const vector< vector< pair<unsigned int, unsigned int> > >& feature_occ;
  
  typedef 
  map< vector< pair< int, int> >, int >  
  production_map_t;
  
  production_map_t& production_map;

  double  (Kernel::*_getKernel)(const feature_node *, const feature_node *) const;
  double  (Kernel::*_getKernel2)(const double) const;

  // J.K to = k(x,e1), k(x,e2) ... k(x,eL)
  void  (Kernel::*_getKernel3)(const feature_node*   x,
                               feature_node**  const xs,
                               unsigned int          size,
                               double*               dest ) const; 
  
  inline void _getKernel_default3( const feature_node*   x,
                                   feature_node**  const xs,
                                   unsigned int          size,
                                   double*               dest  ) const 
  {
    for( unsigned int i = 0; i < size; i++ ) {
      dest[i] =  (this->*_getKernel)( x, xs[i] );
    }
  }

  // linear
  inline double _getKernel_linear(const feature_node *x1, const feature_node *x2) const
  { 
    return dot_normal(x1,x2);
  }

  inline double _getKernel_linear2(const double _x) const
  { 
    return _x; 
  }

  // polynomial
  inline double _getKernel_poly(const feature_node *x1, const feature_node *x2) const 
  { 
    return pow (param_s * dot_normal(x1,x2) + param_r, degree); 
  }

  inline double _getKernel_poly2(const double _x) const 
  { 
    return pow (param_s * _x + param_r, degree); 
  }

  // neural
  inline double _getKernel_neural(const feature_node *x1, const feature_node *x2) const 
  { 
    return tanh (param_s * dot_normal(x1,x2) + param_r); 
  }

  inline double _getKernel_neural2(const double _x) const
  { 
    return tanh (param_s * _x + param_r); 
  }

  // RBF
  inline double _getKernel_rbf(const feature_node *x1, const feature_node *x2) const 
  { 
    return exp (-param_s * norm2(x1,x2));
  }

  inline double _getKernel_rbf2(const double _x) const
  { 
    fprintf (stderr, "Kernel::getKernel_rbf() cannot obtain kernel value only with dot.\n");
    exit (-1);
    return 0.0;
  }

  // ANOVA
  inline double _getKernel_anova(const feature_node *x1, const feature_node *x2) const 
  { 
    register double sum = 0;
    register int zero = d;

    while (x1->index >= 0 && x2->index >= 0) {
      if (x1->index == x2->index) {
	sum += exp (-param_s * (x1->value - x2->value) * (x1->value - x2->value));
	++x1; ++x2;
      } else if (x1->index < x2->index) {
	sum += exp (-param_s * x1->value * x1->value);
	++x1;
      } else {
	sum += exp (-param_s * x2->value * x2->value);
	++x2;
      }
      zero--;
    }

    return pow (sum + (double)zero, degree);
  }

  inline double _getKernel_anova2(const double _x) const
  { 
    fprintf (stderr, "Kernel::getKernel_anova() cannot obtain kernel value only with dot.\n");
    exit (-1);
    return 0.0;
  }

  inline double norm2(const feature_node *x1, const feature_node *x2) const
  {
    register double sum = 0;
     
    while (x1->index >= 0 && x2->index >= 0) {
      if(x1->index == x2->index) {
	sum += (x1->value - x2->value) * (x1->value - x2->value);
	++x1; ++x2;
      } else if (x1->index < x2->index) {
	sum += (x1->value * x1->value);
	++x1;
      } else {
	sum += (x2->value * x2->value);
	++x2;
      }
    }

    while (x1->index >= 0) {
      sum += (x1->value * x1->value);
      ++x1;
    };

    while (x2->index >= 0) {
      sum += (x2->value * x2->value);
      ++x2;
    };
     
    return sum;
  }

  inline double dot_normal(const feature_node *x1, const feature_node *x2) const
  {
    register double sum = 0;
    if ( feature_weight.size() > 0 ) {
      while (x1->index >= 0 && x2->index >= 0) {
        if (x1->index == x2->index) {
	sum += (feature_weight[x1->index] * x1->value * x2->value);
	++x1; ++x2;
        } else if (x1->index < x2->index) {
          ++x1;
        }	else {
          ++x2;
        }			
      }
    } else {
      while (x1->index >= 0 && x2->index >= 0) {
        if (x1->index == x2->index) {
	sum += (x1->value * x2->value);
	++x1; ++x2;
        } else if (x1->index < x2->index) {
          ++x1;
        }	else {
          ++x2;
        }			
      }
    }
    return sum;
  }

  inline int dot_binary(const feature_node *x1, const feature_node *x2) const
  {
    register int sum = 0;
    while (x1->index >= 0 && x2->index >= 0) {
      if (x1->index == x2->index) {
	sum++; 
	++x1; ++x2;
      } else if (x1->index < x2->index) {
	++x1;
      } else {
	++x2;
      }			
    }
    return sum;
  }

  // TREE UTILITY
  static inline double** _make_matrix( int n, int m )
  {
    double** ret;
    ret = new double*[n];
    for(int i = 0; i < n; i++){
      ret[i] = new double[m];
    }
    return ret;
  }

  static inline void _free_matrix( double** mat, int n, int m )
  {
    for(int i = 0; i < n; i++){
      delete[] mat[i];
    }
    delete[] mat;
  }
  
  inline void _resize_tree_mat( double*** mat, int old_size, int new_size ) const
  {
    if ( *mat ) { _free_matrix( *mat, old_size, old_size); }
    *mat = _make_matrix( new_size, new_size );
  }

  // TREE
  inline double _getKernel_tree_naive(const feature_node *x1, const feature_node *x2) const {
    if ( verbose > 0 ) { cerr << "_getKernel_tree_naive\n" << flush; }
    if ( !x1->extra || !x2->extra ) return 0.0;
    
    const LOTree* t1 = x1->extra->getTree();
    const LOTree* t2 = x2->extra->getTree();
    
    if ( !t1 || !t2 ) return 0.0;

    vector< string > t1_sv;
    vector< string > t2_sv;
      
    t1->getAllSubtreeStringRec( t1_sv, false, false ); // ignore mark/geo
    t2->getAllSubtreeStringRec( t2_sv, false, false ); // ignore mark/geo

    map< string, int > t1_sm;
    map< string, int > t2_sm;
    for( unsigned int i = 0; i < t1_sv.size(); i++ ) {
      string& s = t1_sv[ i ]; 
      if ( t1_sm.find( s ) != t1_sm.end() ) {
        t1_sm[ s ] += 1;
      } else {
        t1_sm[ s ] = 1;
      }
    }
    for( unsigned int i = 0; i < t2_sv.size(); i++ ) {
      string& s = t2_sv[ i ]; 
      if ( t2_sm.find( s ) != t2_sm.end() ) {
        t2_sm[ s ] += 1;
      } else {
        t2_sm[ s ] = 1;
      }
    }
    double sum = 0.0;
    for ( map< string, int >::const_iterator itr = t1_sm.begin();
          itr != t1_sm.end();
          itr++ ) {
      const string& s = itr->first;
      LOTree st( s ); //
      double weight = pow( depth_factor, st.numNodes() );
      if ( t2_sm.find( s ) != t2_sm.end() ) {
        double v1 = t1_sm[ s ]; double v2 = t2_sm[ s ];
        sum += weight * v1 * v2;
        if ( verbose > 0 ) {
          cerr << "subtree:\n" << s << "\n";
          cerr << "#1=" << v1 << " #2=" << v2 << "\n";
        }
      }
    }
    return sum;
  }
  
  inline double _getKernel_tree(const feature_node *x1, const feature_node *x2) const 
  { 
    if ( verbose > 0 ) { cerr << "_getKernel_tree\n" << flush; }
    if ( !x1->extra || !x2->extra ) return 0.0;
    
    const LOTree* t1 = x1->extra->getTree();
    const LOTree* t2 = x2->extra->getTree();
    
    if ( !t1 || !t2 ) return 0.0;
    int nt1 = t1->numNodes();
    int nt2 = t2->numNodes();
    int m = max(nt1, nt2);
    if ( m > max_tree_size ) {
      _resize_tree_mat(&tree_c, max_tree_size, m ); 
      _resize_tree_mat(&tree_s,  max_tree_size, m ); // ch_num <= tree_size - 1
      cerr << "resize matrix from  " 
                << max_tree_size << " to " << m << "\n" << flush;
      max_tree_size = m;
    }
      
    if ( verbose > 0 ) {
      cerr << "# nodes of t1 = " << nt1 << "\n" << flush;
      cerr << "# nodes of t2 = " << nt2 << "\n" << flush;
    }

    for ( int n1 = 0; n1 < nt1; n1++ ) {
      const LOTreeNode* node1 = t1->getNodePostOrdered( n1 );
      int nc1 = node1->numChildren();
      for ( int n2 = 0; n2 < nt2; n2++ ) {
        const LOTreeNode* node2 = t2->getNodePostOrdered( n2 );
        int nc2 = node2->numChildren();
          
        if ( node1->isLeaf() && node2->isLeaf() ) {
          if ( node1->getLabel() != node2->getLabel() ) { 
            tree_c[ n1 ][ n2 ] = 0.0;
          } else {
            tree_c[ n1 ][ n2 ] = depth_factor;
          }
        } else if ( node1->getLabel() != node2->getLabel() ) {
          tree_c[ n1 ][ n2 ] = 0.0;
        } else { // not being a leaf and having the same label
          for ( int j = 0; j <= nc2; j++ ) { tree_s[ 0 ][ j ] = 1.0; }
          for ( int i = 0; i <= nc1; i++ ) { tree_s[ i ][ 0 ] = 1.0; } 
          int i = 1;
          const LOTreeNode* ch1 = node1->getFirstChild();
          while ( ch1 ) {
            int ch1_no = ch1->getNoPostOrdered();
            int j = 1;
            const LOTreeNode* ch2 = node2->getFirstChild();
            while ( ch2 ) {
              int ch2_no = ch2->getNoPostOrdered();
              tree_s[ i ][ j ] 
                = tree_s[ i-1 ][ j ] + tree_s[ i ][ j-1 ] 
                - tree_s[ i-1 ][ j-1 ] 
                + tree_s[ i-1 ][ j-1 ] * tree_c[ ch1_no ][ ch2_no ];
              ch2 = ch2->getNextSibling(); j++;
            }
            ch1 = ch1->getNextSibling(); i++;
          }
          tree_c[ n1 ][ n2 ] = depth_factor*tree_s[ nc1 ][ nc2 ];
        }
      }
    }
    double sum  = 0.0;
    for ( int n1 = 0; n1 < nt1; n1++ ) {
      for ( int n2 = 0; n2 < nt2; n2++ ) {
        sum += tree_c[ n1 ][ n2 ];
      }
    }
    if ( verbose > 0 ) {
      cerr << "tree1:\n" << t1->toString() << "\n";
      cerr << "tree2:\n" << t2->toString() << "\n";
      cerr << "kernel = " << sum << "\n" << flush;
    }

    if ( tree_check ) {
      double nv = _getKernel_tree_naive( x1, x2 );
      if ( sum != nv ) {
        cerr << "returning " << sum << ", although naive returns " << nv << "\n";
        exit( 1 );
      }
    }
    return sum;
  }

  inline double _getKernel_unknown(const feature_node *x1, const feature_node *x2) const 
  {
    fprintf (stderr, "Unsupported Kernel. (please implement)\n" );
    exit (-1);
    return 0.0;
  }

  inline double _getKernel_unknown(const double _x) const
  { 
    fprintf (stderr, "Unsupported Kernel. (please implement)n");
    exit (-1);
    return 0.0;
  }
  
  inline double _getKernel_invalid2(const double _x) const
  { 
    fprintf (stderr, "cannot obtain kernel value only with dot.\n");
    exit (-1);
    return 0.0;
  }


  // TREE RESTRICTED
  inline double _getKernel_tree_restrict1_naive(const feature_node *x1, const feature_node *x2) const {
    if ( verbose > 0 ) { cerr << "_getKernel_tree_restrict1_naive\n" << flush; }
    if ( !x1->extra || !x2->extra ) return 0.0; 

    const LOTree* t1 = x1->extra->getTree();
    const LOTree* t2 = x2->extra->getTree();
    if ( !t1 || !t2 ) return 0.0;

    vector< string > t1_sv;
    vector< string > t2_sv;
      
    t1->getAllSubtreeStringRec( t1_sv, true, false ); // ignore geo
    t2->getAllSubtreeStringRec( t2_sv, true, false ); // ignore geoXS

    map< string, int > t1_sm;
    map< string, int > t2_sm;
    for( unsigned int i = 0; i < t1_sv.size(); i++ ) {
      string& s = t1_sv[ i ];
      LOTree st( s ); //
      if ( st.containAtLeastOneMarkRec() ) { //
        if ( t1_sm.find( s ) != t1_sm.end() ) {
          t1_sm[ s ] += 1;
        } else {
          t1_sm[ s ] = 1;
        }
      }
    }
    for( unsigned int i = 0; i < t2_sv.size(); i++ ) {
      string& s = t2_sv[ i ]; 
      LOTree st( s ); //
      if ( st.containAtLeastOneMarkRec() ) { //
        if ( t2_sm.find( s ) != t2_sm.end() ) {
          t2_sm[ s ] += 1;
        } else {
          t2_sm[ s ] = 1;
        }
      }
    }
    double sum = 0.0;
    for ( map< string, int >::const_iterator itr = t1_sm.begin();
          itr != t1_sm.end();
          itr++ ) {
      const string& s = itr->first;
      LOTree st( s ); //
      double weight = pow( depth_factor, st.numNodes() );
      if ( t2_sm.find( s ) != t2_sm.end() ) {
        double v1 = t1_sm[ s ]; double v2 = t2_sm[ s ];
        sum += weight * v1 * v2;
        if ( verbose > 0 ) {
          cerr << "subtree:\n" << s << "\n";
          cerr << "#1=" << v1 << " #2=" << v2 << "\n";
        }
      }
    }
    return sum;
  }

  inline 
  bool can_tree_vec_dot( const feature_node *x1, 
                         const feature_node *x2 ) const
  {
    if (x1->extra->getTreeVec() && x2->extra->getTreeVec() ) {
      if ( ignore_malicious ||
           ( x1 != x2 && 
             !x1->extra->isMaliciousPair( x2->extra->getNo() )) ) {
        return true;
      } else {
        return false;
      }
    } else {
      return false;
    }
  }

  inline 
  double 
  _getKernel_tree_restrict1(const feature_node *x1, 
                            const feature_node *x2) const 
  {
    if ( !x1->extra || !x2->extra ) return 0.0; 

    // malicious-speed-up
    if ( can_tree_vec_dot( x1, x2 ) ) {
      return dot_normal( x1->extra->getTreeVec(),
                         x2->extra->getTreeVec() );
    }

    const LOTree* t1 = x1->extra->getTree();
    const LOTree* t2 = x2->extra->getTree();

    double sum =  _getKernel_tree_restrict1_bare( t1, t2 );
    if ( tree_check ) {
      double nv = _getKernel_tree_restrict1_naive( x1, x2);
      if ( sum != nv ) {
        cerr << "returning " << sum << ", although naive returns " << nv << "\n";
        exit( 1 );
      }
    }

    if ( threshold_normalize == THRESHOLD_MALICIOUS_NORMALIZE ) {
      if (x1->extra->getTreeVec() && x2->extra->getTreeVec() ) {
        double d = dot_normal( x1->extra->getTreeVec(),
                               x2->extra->getTreeVec() );
        //cerr << "malicious normalize" << flush;
        return (threshold_normalize_threshold * ( sum - d ) + d);
      } else {
        cerr << "Can't apply THRESHOLD_MALICIOUS_NORMALIZE\n" << flush;
        exit(1);
      }
    }

    return sum;

    /*
    if ( restriction_ratio < 1.0 ) {
      double no_rest = _getKernel_tree( x1, x2 );
      double ret = (1.0-restriction_ratio)*no_rest + restriction_ratio*sum;
      if ( verbose > 0 ) {
        cerr << (1.0-restriction_ratio) << "*" << no_rest << " + " << restriction_ratio << "*" << sum << " = " << ret << "\n" << flush;
      }
      return ret;
    } else {
      return sum;
    }
    */
  }

  inline 
  double 
  _getKernel_tree_restrict1_bare(const LOTree *t1, 
                                 const LOTree *t2) const 
  { 
    // the meaning of c and s is slightly different from 
    // that of in _getKernel_tree
    // Here, c and s stores the values for marking-conformant subtrees.
    if ( verbose > 0 ) {
      cerr << "_getKernel_tree_restrict1\n" << flush;
    }
       
    if ( !t1 || !t2 ) return 0.0;    
      
    int nt1 = t1->numNodes();
    int nt2 = t2->numNodes();
    int m = max(nt1, nt2);

    if ( m > max_tree_size ) {
      _resize_tree_mat(&tree_c,  max_tree_size, m ); 
      _resize_tree_mat(&tree_s,  max_tree_size, m ); // ch_num <= tree_size - 1
      _resize_tree_mat(&tree_cr, max_tree_size, m );
      _resize_tree_mat(&tree_sr, max_tree_size, m );
      if ( verbose > 0 ) {
        cerr << "resize matrix from  " 
                  << max_tree_size << " to " << m << "\n" << flush;
      }
      max_tree_size = m;
    }

    if ( verbose > 0 ) {
      cerr << "# nodes of t1 = " << nt1 << "\n" << flush;
      cerr << "# nodes of t2 = " << nt2 << "\n" << flush;
    }

    /*
      for ( int n1 = 0; n1 < nt1; n1++ ) {
      for ( int n2 = 0; n2 < nt2; n2++ ) {
      tree_c[ n1 ][ n2 ]  = 0.0;
      tree_cr[ n1 ][ n2 ] = 0.0;
      }
      }
    */

    for ( int n1 = 0; n1 < nt1; n1++ ) {
      const LOTreeNode* node1 = t1->getNodePostOrdered( n1 );
      int nc1 = node1->numChildren();
      for ( int n2 = 0; n2 < nt2; n2++ ) {
        const LOTreeNode* node2 = t2->getNodePostOrdered( n2 );
        int nc2 = node2->numChildren();
          
        if ( node1->isLeaf() && node2->isLeaf() ) {
          if ( node1->getLabel() != node2->getLabel() 
               || node1->getMark() != node2->getMark() ) { 
            tree_c[ n1 ][ n2 ] = 0.0; tree_cr[ n1 ][ n2 ] = 0.0;
          } else {
            tree_c[ n1 ][ n2 ] = depth_factor;
            if ( node1->isMarked() && node2->isMarked() ) {
              tree_cr[ n1 ][ n2 ] = depth_factor;
            } else {
              tree_cr[ n1 ][ n2 ] = 0.0;
            }
          }
        } else if ( node1->getLabel() != node2->getLabel() 
                    || node1->getMark() != node2->getMark() ) {
          tree_c[ n1 ][ n2 ]  = 0.0;
          tree_cr[ n1 ][ n2 ] = 0.0;
        } else { // not being a leaf and having the same label and marking
          for ( int j = 0; j <= nc2; j++ ) { tree_s[ 0 ][ j ] = 1.0; }
          if ( node1->isMarked() && node2->isMarked() ) {
            for ( int j = 0; j <= nc2; j++ ) { tree_sr[ 0 ][ j ] = 1.0; }
          } else {
            for ( int j = 0; j <= nc2; j++ ) { tree_sr[ 0 ][ j ] = 0.0; }
          }
          for ( int i = 0; i <= nc1; i++ ) { tree_s[ i ][ 0 ] = 1.0; } 
          if ( node1->isMarked() && node2->isMarked() ) {
            for ( int i = 0; i <= nc1; i++ ) { tree_sr[ i ][ 0 ] = 1.0; }
          } else {
            for ( int i = 0; i <= nc1; i++ ) { tree_sr[ i ][ 0 ] = 0.0; }
          }
          int i = 1;
          const LOTreeNode* ch1 = node1->getFirstChild();
          while ( ch1 ) {
            int ch1_no = ch1->getNoPostOrdered();
            int j = 1;
            const LOTreeNode* ch2 = node2->getFirstChild();
            while ( ch2 ) {
              int ch2_no = ch2->getNoPostOrdered();
              tree_s[ i ][ j ] 
                = tree_s[ i-1 ][ j ] + tree_s[ i ][ j-1 ] 
                - tree_s[ i-1 ][ j-1 ] 
                + tree_s[ i-1 ][ j-1 ] * tree_c[ ch1_no ][ ch2_no ];
              tree_sr[ i ][ j ] 
                = tree_sr[ i-1 ][ j ] + tree_sr[ i ][ j-1 ]
                - tree_sr[ i-1 ][ j-1]
                + tree_sr[ i-1 ][ j-1 ] * tree_c[ ch1_no ][ ch2_no ]
                + tree_s[ i-1 ][ j-1 ]  * tree_cr[ ch1_no ][ ch2_no ]
                - tree_sr[ i-1 ][ j-1 ] * tree_cr[ ch1_no ][ ch2_no ];
              ch2 = ch2->getNextSibling(); j++;
            }
            ch1 = ch1->getNextSibling(); i++;
          }
          tree_c[ n1 ][ n2 ]  = depth_factor*tree_s[ nc1 ][ nc2 ];
          tree_cr[ n1 ][ n2 ] = depth_factor*tree_sr[ nc1 ][ nc2 ];
        }
      }
    }
    double sum  = 0.0;
    for ( int n1 = 0; n1 < nt1; n1++ ) {
      for ( int n2 = 0; n2 < nt2; n2++ ) {
        sum += tree_cr[ n1 ][ n2 ];
      }
    }
    if ( verbose > 0 ) {
      cerr << "tree1:\n" << t1->toString() << "\n";
      cerr << "tree2:\n" << t2->toString() << "\n";
      cerr << "kernel = " << sum << "\n" << flush;
    }
    return sum;
  }

  inline 
  double 
  _getKernel_tree_restrict1_fast(const feature_node *x1, 
                                 const feature_node *x2) const 
  {
    if ( !x1->extra || !x2->extra ) return 0.0; 
    
    // malicious-speed-up
    if ( can_tree_vec_dot( x1, x2 ) ) {
      return dot_normal( x1->extra->getTreeVec(),
                         x2->extra->getTreeVec() );
    }
    
    const LOTree* t1 = x1->extra->getTree();
    const LOTree* t2 = x2->extra->getTree();
    
    double sum = _getKernel_tree_restrict1_fast_bare( t1, t2 );
 
    if ( tree_check ) {
      double nv = _getKernel_tree_restrict1_naive( x1, x2);
      if ( sum != nv ) {
        cerr << "returning " << sum << ", although naive returns " << nv << "\n";
        exit( 1 );
      }
    }

    if ( threshold_normalize == THRESHOLD_MALICIOUS_NORMALIZE ) {
      if (x1->extra->getTreeVec() && x2->extra->getTreeVec() ) {
        double d = dot_normal( x1->extra->getTreeVec(),
                               x2->extra->getTreeVec() );
        //cerr << "malicious normalize" << flush;
        return (threshold_normalize_threshold * ( sum - d ) + d);
      } else {
        cerr << "Can't apply THRESHOLD_MALICIOUS_NORMALIZE\n" << flush;
        exit(1);
      }
    }

    return sum;
  }

  inline 
  double 
  _getKernel_tree_restrict1_fast_bare(const LOTree *t1, 
                                      const LOTree *t2) const 
  { 
    if ( verbose > 0 ) {
      cerr << "_getKernel_tree_restrict1_fast\n" << flush;
    }

    if ( !t1 || !t2 ) return 0.0;
  
    int nt1 = t1->numNodes();
    int nt2 = t2->numNodes();
    int m = max(nt1, nt2);
    if ( m > max_tree_size ) {
      _resize_tree_mat(&tree_c,  max_tree_size, m ); 
      _resize_tree_mat(&tree_s,  max_tree_size, m ); // ch_num <= tree_size - 1
      _resize_tree_mat(&tree_cr, max_tree_size, m );
      _resize_tree_mat(&tree_sr, max_tree_size, m );
      if ( verbose > 0 ) {
        cerr << "resize matrix from  " 
                  << max_tree_size << " to " << m << "\n" << flush;
      }
      max_tree_size = m;
    }

    if ( verbose > 0 ) {
      cerr << "# nodes of t1 = " << nt1 << "\n" << flush;
      cerr << "# nodes of t2 = " << nt2 << "\n" << flush;
    }

    for ( int n1 = 0; n1 < nt1; n1++ ) {
      for ( int n2 = 0; n2 < nt2; n2++ ) {
        tree_c[ n1 ][ n2 ]  = 0.0;
        tree_cr[ n1 ][ n2 ] = 0.0;
      }
    }

    for ( int n1 = 0; n1 < nt1; n1++ ) {
      const LOTreeNode* node1 = t1->getNodePostOrdered( n1 );
      const vector< int >& snodes 
        = t2->getSameLabelNodes( node1->getLabel() );
      int nc1 = node1->numChildren();
      for ( int k = 0; k < snodes.size(); k++ ) {
        int n2 = snodes[k];
        const LOTreeNode* node2 = t2->getNodePostOrdered( n2 );
        int nc2 = node2->numChildren();
          
        if ( node1->isLeaf() && node2->isLeaf() ) {
          if ( node1->getMark() != node2->getMark() ) { 
            tree_c[ n1 ][ n2 ] = 0.0; tree_cr[ n1 ][ n2 ] = 0.0;
          } else {
            tree_c[ n1 ][ n2 ] = depth_factor;
            if ( node1->isMarked() && node2->isMarked() ) {
              tree_cr[ n1 ][ n2 ] = depth_factor;
            } else {
              tree_cr[ n1 ][ n2 ] = 0.0;
            }
          }
        } else if ( node1->getMark() != node2->getMark() ) {
          tree_c[ n1 ][ n2 ]  = 0.0;
          tree_cr[ n1 ][ n2 ] = 0.0;
        } else { // not being a leaf and having the same label and marking
          for ( int j = 0; j <= nc2; j++ ) { tree_s[ 0 ][ j ] = 1.0; }
          if ( node1->isMarked() && node2->isMarked() ) {
            for ( int j = 0; j <= nc2; j++ ) { tree_sr[ 0 ][ j ] = 1.0; }
          } else {
            for ( int j = 0; j <= nc2; j++ ) { tree_sr[ 0 ][ j ] = 0.0; }
          }
          for ( int i = 0; i <= nc1; i++ ) { tree_s[ i ][ 0 ] = 1.0; } 
          if ( node1->isMarked() && node2->isMarked() ) {
            for ( int i = 0; i <= nc1; i++ ) { tree_sr[ i ][ 0 ] = 1.0; }
          } else {
            for ( int i = 0; i <= nc1; i++ ) { tree_sr[ i ][ 0 ] = 0.0; }
          }
          int i = 1;
          const LOTreeNode* ch1 = node1->getFirstChild();
          while ( ch1 ) {
            int ch1_no = ch1->getNoPostOrdered();
            int j = 1;
            const LOTreeNode* ch2 = node2->getFirstChild();
            while ( ch2 ) {
              int ch2_no = ch2->getNoPostOrdered();
              tree_s[ i ][ j ] 
                = tree_s[ i-1 ][ j ] + tree_s[ i ][ j-1 ] 
                - tree_s[ i-1 ][ j-1 ] 
                + tree_s[ i-1 ][ j-1 ] * tree_c[ ch1_no ][ ch2_no ];
              tree_sr[ i ][ j ] 
                = tree_sr[ i-1 ][ j ] + tree_sr[ i ][ j-1 ]
                - tree_sr[ i-1 ][ j-1]
                + tree_sr[ i-1 ][ j-1 ] * tree_c[ ch1_no ][ ch2_no ]
                + tree_s[ i-1 ][ j-1 ]  * tree_cr[ ch1_no ][ ch2_no ]
                - tree_sr[ i-1 ][ j-1 ] * tree_cr[ ch1_no ][ ch2_no ];
              ch2 = ch2->getNextSibling(); j++;
            }
            ch1 = ch1->getNextSibling(); i++;
          }
          tree_c[ n1 ][ n2 ]  = depth_factor*tree_s[ nc1 ][ nc2 ];
          tree_cr[ n1 ][ n2 ] = depth_factor*tree_sr[ nc1 ][ nc2 ];
        }
      }
    }
    double sum  = 0.0;
    for ( int n1 = 0; n1 < nt1; n1++ ) {
      const LOTreeNode* node1 = t1->getNodePostOrdered( n1 );
      const vector< int >& snodes 
        = t2->getSameLabelNodes( node1->getLabel() );
      for ( int k = 0; k < snodes.size(); k++ ) {
        int n2 = snodes[k];
        sum += tree_cr[ n1 ][ n2 ];
      }
    }
    if ( verbose > 0 ) {
      cerr << "tree1:\n" << t1->toString() << "\n";
      cerr << "tree2:\n" << t2->toString() << "\n";
      cerr << "kernel = " << sum << "\n" << flush;
    }
    
    return sum;
  }

  inline 
  bool 
  _is_same_production_collins_restrict1( const LOTreeNode* x1, 
                                         const LOTreeNode* x2 ) const {
    if ( x1->getLabel() != x2->getLabel() || x1->getMark() != x2->getMark()) {
      return false;
    } else if ( x1->isLeaf() || x2->isLeaf() ) {
      return false;
    } else if ( x1->numChildren() != x2->numChildren() ) {
      return false;
    } else {
      const LOTreeNode* ch1 = x1->getFirstChild();
      const LOTreeNode* ch2 = x2->getFirstChild();
      while( ch1 && ch2 ) {
        if ( ch1->getLabel() != ch2->getLabel() 
             || ch1->getMark() != ch2->getMark()) {
          return false;
        }
        ch1 = ch1->getNextSibling();
        ch2 = ch2->getNextSibling();
      }
      return true;
    }
  }

  inline bool _is_preterminal( const LOTreeNode *x ) const {
    if ( x->numChildren() == 1 && x->getFirstChild()->isLeaf() ) {
      return true;
    } else {
      return false;
    }
  }

  inline 
  double 
  _getKernel_tree_collins_restrict1( const feature_node *x1, 
                                     const feature_node *x2 ) const {
    // direct implementation of the Collins' tree kernel (for restriction)
    if ( verbose > 0 ) {
      cerr << "_getKernel_tree_collins_restrict1\n" << flush;
    }
    if ( !x1->extra || !x2->extra ) return 0.0;

    if ( can_tree_vec_dot( x1, x2 ) ) {
      return dot_normal( x1->extra->getTreeVec(),
                         x2->extra->getTreeVec() );
    }

    const LOTree* t1 = x1->extra->getTree();
    const LOTree* t2 = x2->extra->getTree();
    
    if ( !t1 || !t2 ) return 0.0;

    int nt1 = t1->numNodes();
    int nt2 = t2->numNodes();
    int m = max(nt1, nt2);
    if ( m > max_tree_size ) {
      _resize_tree_mat(&tree_c, max_tree_size, m ); 
      _resize_tree_mat(&tree_cr, max_tree_size, m );
      cerr << "resize matrix from  " 
                << max_tree_size << " to " << m << "\n" << flush;
      max_tree_size = m;
    }

    if ( verbose > 0 ) {
      cerr << "# nodes of t1 = " << nt1 << "\n" << flush;
      cerr << "# nodes of t2 = " << nt2 << "\n" << flush;
    }

    for ( int n1 = 0; n1 < nt1; n1++ ) {
      const LOTreeNode* node1 = t1->getNodePostOrdered( n1 );
      for ( int n2 = 0; n2 < nt2; n2++ ) {
        const LOTreeNode* node2 = t2->getNodePostOrdered( n2 );
        if ( !_is_same_production_collins_restrict1( node1, node2 ) ) {
          tree_c[ n1 ][ n2 ] = 0.0; tree_cr[ n1 ][ n2 ] = 0.0;
        } else if ( _is_preterminal( node1 ) && _is_preterminal( node2 ) ) {
          tree_c[ n1 ][ n2 ] = depth_factor;
          if ( node1->isMarked() && node2->isMarked() ) {
            // assume that the restriction marks are not at leaves
            tree_cr[ n1 ][ n2 ] = depth_factor;
          } else {
            tree_cr[ n1 ][ n2 ] = 0.0;
          }
        } else {
          double c  = 1.0;
          double cr = 0.0;
          if ( node1->isMarked() && node2->isMarked() ) {
            cr = 1.0;
          }
          const LOTreeNode* ch1 = node1->getFirstChild();
          const LOTreeNode* ch2 = node2->getFirstChild();
          while ( ch1 && ch2 ) {
            int ch1_no = ch1->getNoPostOrdered();
            int ch2_no = ch2->getNoPostOrdered();
            double m  = 0.0;
            if ( ch1->isMarked() && ch2->isMarked() ) {
              m  = 1.0;
            }
            double md = 1.0 - m;
            double new_c = c * ( 1.0 + tree_c[ ch1_no ][ ch2_no ] );
            cr = cr * ( md + tree_c[ ch1_no ][ ch2_no ])
              + c * ( m  + tree_cr[ ch1_no ][ ch2_no ])
              - cr * tree_cr[ ch1_no ][ ch2_no ];
            c = new_c;
            ch1 = ch1->getNextSibling();
            ch2 = ch2->getNextSibling();
          }
          // 
          tree_c[ n1 ][ n2 ]  = depth_factor*c;
          tree_cr[ n1 ][ n2 ] = depth_factor*cr;
        }
      }
    }
    double sum  = 0.0;
    for ( int n1 = 0; n1 < nt1; n1++ ) {
      for ( int n2 = 0; n2 < nt2; n2++ ) {
        sum += tree_cr[ n1 ][ n2 ];
      }
    }
    if ( verbose > 0 ) {
      cerr << "tree1:\n" << t1->toString() << "\n";
      cerr << "tree2:\n" << t2->toString() << "\n";
      cerr << "kernel = " << sum << "\n" << flush;
    }
    return sum;
  }

  inline double 
  _getKernel_tree_collins_restrict1_fast( const feature_node *x1, 
                                          const feature_node *x2 ) const {
    // direct implementation of the Collins' tree kernel (for restriction)
    if ( verbose > 0 ) {
      cerr << "_getKernel_tree_collins_restrict1_fast\n" << flush;
    }
    if ( !x1->extra || !x2->extra ) return 0.0; 

    if ( can_tree_vec_dot( x1, x2 ) ) {
      return dot_normal( x1->extra->getTreeVec(),
                         x2->extra->getTreeVec() );
    }

    const LOTree* t1 = x1->extra->getTree();
    const LOTree* t2 = x2->extra->getTree();
    
    if ( !t1 || !t2 ) return 0.0;
  
    int nt1 = t1->numNodes();
    int nt2 = t2->numNodes();
    int m = max(nt1, nt2);
    if ( m > max_tree_size ) {
      _resize_tree_mat(&tree_c, max_tree_size, m ); 
      _resize_tree_mat(&tree_cr, max_tree_size, m );
      if ( verbose > 0 ) {
        cerr << "resize matrix from  " 
                  << max_tree_size << " to " << m << "\n" << flush;
      }
      max_tree_size = m;
    }

    if ( verbose > 0 ) {
      cerr << "# nodes of t1 = " << nt1 << "\n" << flush;
      cerr << "# nodes of t2 = " << nt2 << "\n" << flush;
    }

    for ( int n1 = 0; n1 < nt1; n1++ ) {
      for ( int n2 = 0; n2 < nt2; n2++ ) {
        tree_c[ n1 ][ n2 ]  = 0.0;
        tree_cr[ n1 ][ n2 ] = 0.0;
      }
    }

    for ( int n1 = 0; n1 < nt1; n1++ ) {
      const LOTreeNode* node1 = t1->getNodePostOrdered( n1 );
      const vector< int >& snodes = t2->getSameLabelNodes( node1->getLabel() );
      int nc1 = node1->numChildren();
      for ( int k = 0; k < snodes.size(); k++ ) {
        int n2 = snodes[k];
        const LOTreeNode* node2 = t2->getNodePostOrdered( n2 );
        if ( !_is_same_production_collins_restrict1( node1, node2 ) ) {
          tree_c[ n1 ][ n2 ] = 0.0; tree_cr[ n1 ][ n2 ] = 0.0;
        } else if ( _is_preterminal( node1 ) && _is_preterminal( node2 ) ) {
          tree_c[ n1 ][ n2 ] = depth_factor;
          if ( node1->isMarked() && node2->isMarked() ) {
            // assume that the restriction marks are not at leaves
            tree_cr[ n1 ][ n2 ] = depth_factor;
          } else {
            tree_cr[ n1 ][ n2 ] = 0.0;
          }
        } else {
          double c  = 1.0;
          double cr = 0.0;
          if ( node1->isMarked() && node2->isMarked() ) {
            cr = 1.0;
          }
          const LOTreeNode* ch1 = node1->getFirstChild();
          const LOTreeNode* ch2 = node2->getFirstChild();
          while ( ch1 && ch2 ) {
            int ch1_no = ch1->getNoPostOrdered();
            int ch2_no = ch2->getNoPostOrdered();
            double m  = 0.0;
            if ( ch1->isMarked() && ch2->isMarked() ) {
              m  = 1.0;
            }
            double md = 1.0 - m;
            double new_c = c * ( 1.0 + tree_c[ ch1_no ][ ch2_no ] );
            cr = cr * ( md + tree_c[ ch1_no ][ ch2_no ])
              + c * ( m  + tree_cr[ ch1_no ][ ch2_no ])
              - cr * tree_cr[ ch1_no ][ ch2_no ];
            c = new_c;
            ch1 = ch1->getNextSibling();
            ch2 = ch2->getNextSibling();
          }
          // 
          tree_c[ n1 ][ n2 ]  = depth_factor*c;
          tree_cr[ n1 ][ n2 ] = depth_factor*cr;
        }
      }
    }
    double sum  = 0.0;
     
    for ( int n1 = 0; n1 < nt1; n1++ ) {
      const LOTreeNode* node1 = t1->getNodePostOrdered( n1 );
      const vector< int >& snodes = t2->getSameLabelNodes( node1->getLabel() );
      for ( int k = 0; k < snodes.size(); k++ ) {
        int n2 = snodes[k];
        sum += tree_cr[ n1 ][ n2 ];
      }
    }

    if ( verbose > 0 ) {
      cerr << "tree1:\n" << t1->toString() << "\n";
      cerr << "tree2:\n" << t2->toString() << "\n";
      cerr << "kernel = " << sum << "\n" << flush;
    }
    return sum;
  }
  
  inline double 
  _getKernel_tree_collins_p_restrict1( const feature_node *x1, 
                                       const feature_node *x2 ) const {
    if ( verbose > 0 ) {
      cerr << "_getKernel_tree_collins_p_restrict1\n" << flush;
    }
    if ( !x1->extra || !x2->extra ) return 0.0;
    
    // malicious-speed-up
    if ( can_tree_vec_dot( x1, x2 ) ) {
      return dot_normal( x1->extra->getTreeVec(),
                         x2->extra->getTreeVec() );
    }
    // delegate
    return 
      _getKernel_tree_restrict1_bare( x1->extra->getTreeCollinsP(),
                                      x2->extra->getTreeCollinsP() );
  }

  inline double 
  _getKernel_tree_collins_p_restrict1_fast( const feature_node *x1, 
                                            const feature_node *x2 ) const {
    if ( verbose > 0 ) {
      cerr << "_getKernel_tree_collins_p_restrict1_fast\n" << flush;
    }
    if ( !x1->extra || !x2->extra ) return 0.0; 

    if ( can_tree_vec_dot( x1, x2 ) ) {
      return dot_normal( x1->extra->getTreeVec(),
                         x2->extra->getTreeVec() );
    }

    // delegate
    return 
      _getKernel_tree_restrict1_fast_bare( x1->extra->getTreeCollinsP(),
                                           x2->extra->getTreeCollinsP() );
  }

  

  inline 
  double 
  _getKernel_tree_collins_plus_restrict1(const feature_node *x1, 
                                         const feature_node *x2 ) const 
  { 
    if ( verbose > 0 ) {
      cerr << "_getKernel_tree_collins_plus_restrict1\n" << flush;
    }
    if ( !x1->extra || !x2->extra ) return 0.0; 
    const LOTree* t1 = x1->extra->getTree();
    const LOTree* t2 = x2->extra->getTree();
    if ( !t1 || !t2 ) return 0.0;

    int nt1 = t1->numNodes();
    int nt2 = t2->numNodes();
    int m = max(nt1, nt2);
    if ( m > max_tree_size ) {
      _resize_tree_mat(&tree_c, max_tree_size, m ); 
      _resize_tree_mat(&tree_cr, max_tree_size, m );
      if ( verbose > 0 ) {
        cerr << "resize matrix from  " 
                  << max_tree_size << " to " << m << "\n" << flush;
      }
      max_tree_size = m;
    }

    if ( verbose > 0 ) {
      cerr << "# nodes of t1 = " << nt1 << "\n" << flush;
      cerr << "# nodes of t2 = " << nt2 << "\n" << flush;
    }

    for ( int n1 = 0; n1 < nt1; n1++ ) {
      const LOTreeNode* node1 = t1->getNodePostOrdered( n1 );
      int nc1 = node1->numChildren();
      for ( int n2 = 0; n2 < nt2; n2++ ) {
        const LOTreeNode* node2 = t2->getNodePostOrdered( n2 );
        int nc2 = node2->numChildren();
          
        if ( node1->isLeaf() && node2->isLeaf() ) {
          if ( node1->getLabel() != node2->getLabel() 
               || node1->getMark() != node2->getMark() ) { 
            tree_c[ n1 ][ n2 ] = 0.0; tree_cr[ n1 ][ n2 ] = 0.0;
          } else {
            tree_c[ n1 ][ n2 ] = depth_factor;
            if ( node1->isMarked() && node2->isMarked() ) {
              tree_cr[ n1 ][ n2 ] = depth_factor;
            } else {
              tree_cr[ n1 ][ n2 ] = 0.0;
            }
          }
        } else if ( node1->getLabel() != node2->getLabel() 
                    || node1->getMark() != node2->getMark() ) {
          tree_c[ n1 ][ n2 ]  = 0.0;
          tree_cr[ n1 ][ n2 ] = 0.0;
        } else { // not being a leaf and having the same label and marking

          if ( nc1 != nc2 ) {
            tree_c[ n1  ][ n2 ] = depth_factor;
            if ( node1->isMarked() && node2->isMarked() ) {
              tree_cr[ n1 ][ n2 ] = depth_factor;
            } else {
              tree_cr[ n1 ][ n2 ] = 0.0;
            }
          }
          else {
            double c  = 1.0;
            double cr = 0.0;
            if ( node1->isMarked() && node2->isMarked() ) {
              cr = 1.0;
            }
            const LOTreeNode* ch1 = node1->getFirstChild();
            const LOTreeNode* ch2 = node2->getFirstChild();
            while ( ch1 && ch2 ) {
              int ch1_no = ch1->getNoPostOrdered();
              int ch2_no = ch2->getNoPostOrdered();
                
              double new_c = c*tree_c[ ch1_no ][ ch2_no ];
              cr = cr * tree_c[ ch1_no ][ ch2_no ]
                + c * tree_cr[ ch1_no ][ ch2_no ]
                - cr * tree_cr[ ch1_no ][ ch2_no ];
              c = new_c;
              ch1 = ch1->getNextSibling();
              ch2 = ch2->getNextSibling();
            }
            // 
            tree_c[ n1 ][ n2 ]  = depth_factor*(c+1.0);
            tree_cr[ n1 ][ n2 ] = depth_factor*cr;
            if ( node1->isMarked() && node2->isMarked() ) {
              tree_cr[ n1 ][ n2 ] += depth_factor;
            }
          }
        }
      }
    }
    double sum  = 0.0;
    for ( int n1 = 0; n1 < nt1; n1++ ) {
      for ( int n2 = 0; n2 < nt2; n2++ ) {
        sum += tree_cr[ n1 ][ n2 ];
      }
    }
    if ( verbose > 0 ) {
      cerr << "tree1:\n" << t1->toString() << "\n";
      cerr << "tree2:\n" << t2->toString() << "\n";
      cerr << "kernel = " << sum << "\n" << flush;
    }
    return sum;
  }


  // TREE_ELASTIC
  inline 
  double 
  _getKernel_tree_elastic_naive(const feature_node *x1, 
                                const feature_node *x2) const {
    if ( verbose > 0 ) { 
      cerr << "_getKernel_tree_elastic_naive\n" << flush; 
    }
    if ( !x1->extra || !x2->extra )  return 0.0;
    const LOTree* t1 = x1->extra->getTree();
    const LOTree* t2 = x2->extra->getTree();
    if ( !t1 || !t2 ) return 0.0;

    vector< string > t1_sv;
    vector< string > t2_sv;
      
    t1->getAllElasticSubtreeStringRec( t1_sv, 
                                       false, false ); // ignore mark/geo
    t2->getAllElasticSubtreeStringRec( t2_sv, 
                                       false, false ); // ignore mark/geo

    map< string, int > t1_sm;
    map< string, int > t2_sm;
    for( unsigned int i = 0; i < t1_sv.size(); i++ ) {
      string& s = t1_sv[ i ]; 
      if ( t1_sm.find( s ) != t1_sm.end() ) {
        t1_sm[ s ] += 1;
      } else {
        t1_sm[ s ] = 1;
      }
    }
    for( unsigned int i = 0; i < t2_sv.size(); i++ ) {
      string& s = t2_sv[ i ]; 
      if ( t2_sm.find( s ) != t2_sm.end() ) {
        t2_sm[ s ] += 1;
      } else {
        t2_sm[ s ] = 1;
      }
    }
    double sum = 0.0;
    for ( map< string, int >::const_iterator itr = t1_sm.begin();
          itr != t1_sm.end();
          itr++ ) {
      const string& s = itr->first;
      LOTree st( s ); //
      double weight = pow( depth_factor, st.numNodes() );
      if ( t2_sm.find( s ) != t2_sm.end() ) {
        double v1 = t1_sm[ s ]; double v2 = t2_sm[ s ];
        sum += weight * v1 * v2;
        if ( verbose > 0 ) {
          cerr << "subtree:\n" << s << "\n";
          cerr << "#1=" << v1 << " #2=" << v2 << "\n";
        }
      }
    }
    return sum;
  }

  inline double 
  _tree_kernel_elastic_calc_ca( double** c, double** ca, int n1, int n2, 
                                const LOTreeNode* node1, 
                                const LOTreeNode* node2 ) const
  {
    double a = 0.0;
    const LOTreeNode* ch1 = node1->getFirstChild();
    while ( ch1 ) { 
      int ch1_no = ch1->getNoPostOrdered();
      a += ca[ ch1_no ][ n2 ];
      ch1 = ch1->getNextSibling();
    }
    const LOTreeNode* ch2 = node2->getFirstChild();
    while ( ch2 ) { 
      int ch2_no = ch2->getNoPostOrdered();
      a += ca[ n1 ][ ch2_no ];
      ch2 = ch2->getNextSibling();
    }
    ch1 = node1->getFirstChild();
    while ( ch1 ) { 
      int ch1_no = ch1->getNoPostOrdered();
      ch2 = node2->getFirstChild();
      while ( ch2 ) { 
        int ch2_no = ch2->getNoPostOrdered();
        a -= ca[ ch1_no ][ ch2_no ];
        ch2 = ch2->getNextSibling();
      }
      ch1 = ch1->getNextSibling();
    }
    a += c[ n1 ][ n2 ];
    return a;
  }

  inline double _getKernel_tree_elastic(const feature_node *x1, 
                                        const feature_node *x2) const 
  { 
    if ( verbose > 0 ) {
      cerr << "_getKernel_tree_elastic\n" << flush;
    }
    if ( !x1->extra || !x2->extra ) return 0.0; 
    const LOTree* t1 = x1->extra->getTree();
    const LOTree* t2 = x2->extra->getTree();
      
    if ( !t1 || !t2 ) return 0.0;

    int nt1 = t1->numNodes();
    int nt2 = t2->numNodes();
    int m = max(nt1,nt2);
    if ( m > max_tree_size ) {
      _resize_tree_mat(&tree_c, max_tree_size, m );
      _resize_tree_mat(&tree_s, max_tree_size, m );
      _resize_tree_mat(&tree_ca, max_tree_size, m );
      if ( verbose > 0 ) {
        cerr << "resize matrix from  " 
                  << max_tree_size << " to " << m << "\n" << flush;
      }
      max_tree_size = m;
    }
      
    if ( verbose > 0 ) {
      cerr << "# nodes of t1 = " << nt1 << "\n" << flush;
      cerr << "# nodes of t2 = " << nt2 << "\n" << flush;
    }
            
    for ( int n1 = 0; n1 < nt1; n1++ ) {
      const LOTreeNode* node1 = t1->getNodePostOrdered( n1 );
      int nc1 = node1->numChildren();
      for ( int n2 = 0; n2 < nt2; n2++ ) {
        const LOTreeNode* node2 = t2->getNodePostOrdered( n2 );
        int nc2 = node2->numChildren();
          
        if ( node1->isLeaf() && node2->isLeaf() ) {
          if ( node1->getLabel() != node2->getLabel() ) {
            tree_c[ n1 ][ n2 ] = 0.0; tree_ca[ n1 ][ n2 ] = 0.0;
          } else {
            tree_c[ n1 ][ n2 ]  = depth_factor; 
            tree_ca[ n1 ][ n2 ] = depth_factor;
          }
        } else {
          if ( node1->getLabel() != node2->getLabel() ) {
            tree_c[ n1 ][ n2 ] = 0.0;
          } else {
            for ( int j = 0; j <= nc2; j++ ) { tree_s[ 0 ][ j ] = 1.0; }
            for ( int i = 0; i <= nc1; i++ ) { tree_s[ i ][ 0 ] = 1.0; }
            int i = 1;
            const LOTreeNode* ch1 = node1->getFirstChild();
            while ( ch1 ) {
              int ch1_no = ch1->getNoPostOrdered();
              int j = 1;
              const LOTreeNode* ch2 = node2->getFirstChild();
              while ( ch2 ) {
                int ch2_no = ch2->getNoPostOrdered();
                tree_s[ i ][ j ] 
                  = tree_s[ i-1 ][ j ] + tree_s[ i ][ j-1 ] 
                  - tree_s[ i-1 ][ j-1 ] 
                  + tree_s[ i-1 ][ j-1 ] * tree_ca[ ch1_no ][ ch2_no ];
                ch2 = ch2->getNextSibling(); j++;
              }
              ch1 = ch1->getNextSibling(); i++;
            }
            tree_c[ n1 ][ n2 ] = depth_factor*tree_s[ nc1 ][ nc2 ];
          }
          // calculate ca[n1][n2]
          tree_ca[ n1 ][ n2 ] = depth_factor*
            _tree_kernel_elastic_calc_ca( tree_c, tree_ca, 
                                          n1, n2, node1, node2 );
        }
      }
    }
    double sum  = 0.0;
    for ( int n1 = 0; n1 < nt1; n1++ ) {
      for ( int n2 = 0; n2 < nt2; n2++ ) {
        sum += tree_c[ n1 ][ n2 ];
      }
    }
    if ( verbose > 0 ) {
      cerr << "tree1:\n" << t1->toString() << "\n";
      cerr << "tree2:\n" << t2->toString() << "\n";
      cerr << "kernel = " << sum << "\n" << flush;
    }
    if ( tree_check ) {
      double nv = _getKernel_tree_elastic_naive( x1, x2);
      if ( sum != nv ) {
        cerr << "returning " << sum 
                  << ", although naive returns " << nv << "\n";
        exit( 1 );
      }
    }
    return sum;
  }

  inline 
  double 
  _getKernel_tree_elastic_restrict1_naive(const feature_node *x1, 
                                          const feature_node *x2) const {
    if ( verbose > 0 ) { 
      cerr << "_getKernel_tree_elastic_restrict1_naive\n" << flush; 
    }
    if ( !x1->extra || !x2->extra ) return 0.0; 

    const LOTree* t1 = x1->extra->getTree();
    const LOTree* t2 = x2->extra->getTree();
    
    if ( !t1 || !t2 ) return 0.0;

    vector< string > t1_sv;
    vector< string > t2_sv;
      
    t1->getAllElasticSubtreeStringRec( t1_sv, 
                                       true, false ); // ignore geo
    t2->getAllElasticSubtreeStringRec( t2_sv, 
                                       true, false ); // ignore geo

    map< string, int > t1_sm;
    map< string, int > t2_sm;
    for( unsigned int i = 0; i < t1_sv.size(); i++ ) {
      string& s = t1_sv[ i ];
        
      LOTree st( s ); //
      if ( st.containAtLeastOneMarkRec() ) { //
        if ( t1_sm.find( s ) != t1_sm.end() ) {
          t1_sm[ s ] += 1;
        } else {
          t1_sm[ s ] = 1;
        }
      }
    }
    for( unsigned int i = 0; i < t2_sv.size(); i++ ) {
      string& s = t2_sv[ i ]; 
      LOTree st( s ); //
      if ( st.containAtLeastOneMarkRec() ) { //
        if ( t2_sm.find( s ) != t2_sm.end() ) {
          t2_sm[ s ] += 1;
        } else {
          t2_sm[ s ] = 1;
        }
      }
    }
    double sum = 0.0;
    for ( map< string, int >::const_iterator itr = t1_sm.begin();
          itr != t1_sm.end();
          itr++ ) {
      const string& s = itr->first;
      LOTree st( s ); //
      double weight = pow( depth_factor, st.numNodes() );
      if ( t2_sm.find( s ) != t2_sm.end() ) {
        double v1 = t1_sm[ s ]; double v2 = t2_sm[ s ];
        sum += weight * v1 * v2;
        if ( verbose > 0 ) {
          cerr << "subtree:\n" << s << "\n";
          cerr << "#1=" << v1 << " #2=" << v2 << "\n";
        }
      }
    }
    return sum;
  }

  inline 
  double _getKernel_tree_elastic_restrict1(const feature_node *x1, 
                                           const feature_node *x2) const 
  { 
    if ( verbose > 0 ) {
      cerr << "_getKernel_tree_elastic_restrict1\n" << flush;
    }
    if ( !x1->extra || !x2->extra ) return 0.0; 

    const LOTree* t1 = x1->extra->getTree();
    const LOTree* t2 = x2->extra->getTree();
    if ( !t1 || !t2 ) return 0.0;
    int nt1 = t1->numNodes();
    int nt2 = t2->numNodes();
    int m = max(nt1,nt2);
    if ( m > max_tree_size ) {
      _resize_tree_mat(&tree_c,   max_tree_size, m );
      _resize_tree_mat(&tree_s,   max_tree_size, m );
      _resize_tree_mat(&tree_ca,  max_tree_size, m );
      _resize_tree_mat(&tree_cr,  max_tree_size, m );
      _resize_tree_mat(&tree_sr,  max_tree_size, m );
      _resize_tree_mat(&tree_car, max_tree_size, m );
      if ( verbose > 0 ) {
        cerr << "resize matrix from  " 
                  << max_tree_size << " to " << m << "\n" << flush;
      }
      max_tree_size = m;
    }
      
    if ( verbose > 0 ) {
      cerr << "# nodes of t1 = " << nt1 << "\n" << flush;
      cerr << "# nodes of t2 = " << nt2 << "\n" << flush;
    }
            
    for ( int n1 = 0; n1 < nt1; n1++ ) {
      const LOTreeNode* node1 = t1->getNodePostOrdered( n1 );
      int nc1 = node1->numChildren();
      for ( int n2 = 0; n2 < nt2; n2++ ) {
        const LOTreeNode* node2 = t2->getNodePostOrdered( n2 );
        int nc2 = node2->numChildren();
          
        if ( node1->isLeaf() && node2->isLeaf() ) {
          if ( node1->getLabel() != node2->getLabel() 
               || node1->getMark() != node2->getMark() ) {
            tree_c[ n1 ][ n2 ] = 0.0; tree_ca[ n1 ][ n2 ] = 0.0;
            tree_cr[ n1 ][ n2 ] = 0.0; tree_car[ n1 ][ n2 ] = 0.0; 
          } else {
            tree_c[ n1 ][ n2 ]  = depth_factor; 
            tree_ca[ n1 ][ n2 ] = depth_factor;
            if ( node1->isMarked() && node2->isMarked() ) {
              tree_cr[ n1 ][ n2 ]  = depth_factor; 
              tree_car[ n1 ][ n2 ] = depth_factor;
            } else {
              tree_cr[ n1 ][ n2 ] = 0.0; tree_car[ n1 ][ n2 ] = 0.0;
            }
          }
        } else {
          if ( node1->getLabel() != node2->getLabel() 
               || node1->getMark() != node2->getMark() ) {
            tree_c[ n1 ][ n2 ]  = 0.0;
            tree_cr[ n1 ][ n2 ] = 0.0;
          } else {
            for ( int j = 0; j <= nc2; j++ ) { tree_s[ 0 ][ j ] = 1.0; }
            if ( node1->isMarked() && node2->isMarked() ) {
              for ( int j = 0; j <= nc2; j++ ) { tree_sr[ 0 ][ j ] = 1.0; }
            } else {
              for ( int j = 0; j <= nc2; j++ ) { tree_sr[ 0 ][ j ] = 0.0; }
            }
            for ( int i = 0; i <= nc1; i++ ) { tree_s[ i ][ 0 ] = 1.0; }
            if ( node1->isMarked() && node2->isMarked() ) {
              for ( int i = 0; i <= nc1; i++ ) { tree_sr[ i ][ 0 ] = 1.0; }
            } else {
              for ( int i = 0; i <= nc1; i++ ) { tree_sr[ i ][ 0 ] = 0.0; }
            }
            int i = 1;
            const LOTreeNode* ch1 = node1->getFirstChild();
            while ( ch1 ) {
              int ch1_no = ch1->getNoPostOrdered();
              int j = 1;
              const LOTreeNode* ch2 = node2->getFirstChild();
              while ( ch2 ) {
                int ch2_no = ch2->getNoPostOrdered();
                tree_s[ i ][ j ] 
                  = tree_s[ i-1 ][ j ] + tree_s[ i ][ j-1 ] 
                  - tree_s[ i-1 ][ j-1 ] 
                  + tree_s[ i-1 ][ j-1 ] * tree_ca[ ch1_no ][ ch2_no ];
                tree_sr[ i ][ j ] 
                  = tree_sr[ i-1 ][ j ] + tree_sr[ i ][ j-1 ]
                  - tree_sr[ i-1 ][ j-1]
                  + tree_sr[ i-1 ][ j-1 ] * tree_ca[ ch1_no ][ ch2_no ]
                  + tree_s[ i-1 ][ j-1 ]  * tree_car[ ch1_no ][ ch2_no ]
                  - tree_sr[ i-1 ][ j-1 ] * tree_car[ ch1_no ][ ch2_no ];
                ch2 = ch2->getNextSibling(); j++;
              }
              ch1 = ch1->getNextSibling(); i++;
            }
            tree_c[ n1 ][ n2 ]  = depth_factor*tree_s[ nc1 ][ nc2 ];
            tree_cr[ n1 ][ n2 ] = depth_factor*tree_sr[ nc1 ][ nc2 ];
          }
          // calculate ca[n1][n2]
          tree_ca[ n1 ][ n2 ] = depth_factor *
            _tree_kernel_elastic_calc_ca( tree_c, tree_ca, 
                                          n1, n2, node1, node2 );
          tree_car[ n1 ][ n2 ] = depth_factor *
            _tree_kernel_elastic_calc_ca( tree_cr, tree_car,
                                          n1, n2, node1, node2 );
        }
      }
    }
    double sum  = 0.0;
    for ( int n1 = 0; n1 < nt1; n1++ ) {
      for ( int n2 = 0; n2 < nt2; n2++ ) {
        sum += tree_cr[ n1 ][ n2 ];
      }
    }
    if ( verbose > 0 ) {
      cerr << "tree1:\n" << t1->toString() << "\n";
      cerr << "tree2:\n" << t2->toString() << "\n";
      cerr << "kernel = " << sum << "\n" << flush;
    }
    if ( tree_check ) {
      double nv = _getKernel_tree_elastic_restrict1_naive( x1, x2);
      if ( sum != nv ) {
        cerr << "returning " << sum << ", although naive returns " << nv << "\n";
        exit( 1 );
      }
    }
    return sum;
  }

  inline void _getKernel_tree_common3( const feature_node*   x, 
                                       feature_node**  const xs,
                                       unsigned int          size,
                                       double*               dest ) const
  {
    if ( x->extra && x->extra->getTreeVec() 
         && feature_weight.size() > 0 
         && feature_occ.size() > 0 ) {


      // There is some unefficiency (although correct)
      // - does not take into account comb-conversion with dev-set
      // - does not take into account support vectors

      //cerr << "hoge\n" << flush;
      const feature_node* tv = x->extra->getTreeVec();
      
      unsigned int abs_size = example.getMaxAbsoluteExampleNum();
      vector< double > v( abs_size );
      for( const feature_node* node = tv; node->index >= 0; node++ ) {
        unsigned int fno = node->index;
        double   val    = node->value;
        double   weight = feature_weight[ fno ];
        const vector< pair< unsigned int, unsigned int > >& occ
          = feature_occ[ fno ];
        for( unsigned int i = 0; i < occ.size(); i++ ) {
          unsigned int eno   = occ[i].first;
          if ( eno < abs_size ) {
            double       onum  = occ[i].second;  
            v[ eno ] += (weight * val * onum );
          }
        }
      }
      
      for ( unsigned int k = 0; k < size; k++ ) {
        const feature_node* x2 = xs[k];
        if ( x2->extra && x2->extra->getTree() 
             && x2->extra->getTreeVec() ) {
          int x2_no = x2->extra->getNo();
          if ( x->extra->isMaliciousPair( x2_no ) ) {
            dest[ k ] =  (this->*_getKernel)( x, x2 );
          } else {
            dest[ k ] = v[ x2_no ];
          }

        } else {
          cerr << "error in _getKernel_tree_common3\n" << flush;
          exit(1);
        }
      }
    } else {
      _getKernel_default3( x, xs, size, dest );
    }
  }
  
 public:
  feature_node ** x;
  double         *y;

  Kernel(const BaseExample& example, const Param& param):
    l(example.l), 
    d(example.d), 
    pack_d(example.pack_d), 
    kernel_type(param.kernel_type),
    feature_type(example.feature_type),
    dot_kernel(param.dot_kernel),
    degree(param.degree),
    param_g(param.param_g),
    param_r(param.param_r),
    param_s(param.param_s),
    // J.K.
    max_tree_size( 0 ),
    tree_c( 0 ),
    tree_s( 0 ),
    tree_ca( 0 ),
    tree_cr( 0 ),
    tree_sr( 0 ),
    tree_car( 0 ),
    verbose(param.verbose),
    tree_check(param.tree_check),
    geometric_type(param.geometric_type),
    node_restriction_type(param.node_restriction_type),
    restriction_ratio( param.restriction_ratio ),
    self_normalize( param.self_normalize ),
    depth_factor( param.depth_factor ),
    threshold_normalize( param.threshold_normalize ),
    threshold_normalize_threshold( param.threshold_normalize_threshold ),
    example( example ),
    feature_weight( *example.feature_weight ),
    feature_occ( *example.feature_occ ),
    production_map( *example.production_map ),
    ignore_malicious( param.ignore_malicious ),
    ignore_trie( param.itrie )
  {
    
    cout << flush;
    if ( self_normalize ) {
      cerr << "do self normalize\n" << flush;
    }

    // default
    switch (kernel_type) {
    case LINEAR:
      _getKernel  = &Kernel::_getKernel_linear;
      _getKernel2 = &Kernel::_getKernel_linear2;
      _getKernel3 = &Kernel::_getKernel_default3;
      break;
    case POLY:
      _getKernel  = &Kernel::_getKernel_poly;
      _getKernel2 = &Kernel::_getKernel_poly2;
      _getKernel3 = &Kernel::_getKernel_default3;
      break;
    case NEURAL:
      _getKernel  = &Kernel::_getKernel_neural;
      _getKernel2 = &Kernel::_getKernel_neural2;
      _getKernel3 = &Kernel::_getKernel_default3;
      break;
    case RBF:
      _getKernel  = &Kernel::_getKernel_rbf;
      _getKernel2 = &Kernel::_getKernel_rbf2;
      _getKernel3 = &Kernel::_getKernel_default3;
      break;
    case ANOVA:
      _getKernel  = &Kernel::_getKernel_anova;
      _getKernel2 = &Kernel::_getKernel_anova2;
      _getKernel3 = &Kernel::_getKernel_default3;
      break;      
    case TREE: // J.K.
      if ( node_restriction_type == NO_RESTRICT ) {
        if ( geometric_type == NO_GEOMETRIC ) {
          _getKernel = &Kernel::_getKernel_tree;
          _getKernel3 = &Kernel::_getKernel_tree_common3;
        } else if ( geometric_type == GEOMETRIC_1 ) {
          _getKernel = &Kernel::_getKernel_unknown;
          _getKernel3 = &Kernel::_getKernel_default3;
          //_getKernel = &Kernel::_getKernel_tree_geo1;
        } else {
          cerr << "Unsupported geometric type: " 
                    << geometric_type << "\n" << flush;
          exit( 1 );
        }
      } else if ( node_restriction_type == RESTRICT_1 ) {
        if ( geometric_type == NO_GEOMETRIC ) {
          _getKernel = &Kernel::_getKernel_tree_restrict1;
          _getKernel3 = &Kernel::_getKernel_tree_common3;
        } else if ( geometric_type == GEOMETRIC_1 ) {
          _getKernel = &Kernel::_getKernel_unknown;
          _getKernel3 = &Kernel::_getKernel_tree_common3;
          //_getKernel = &Kernel::_getKernel_tree_restrict1_geo1;
        } else {
          cerr << "Unsupported geometric type: " 
                    << geometric_type << "\n" << flush;
          exit( 1 );
        }
      } else {
        cerr << "Unsupported restriction type: " 
                  << node_restriction_type << "\n" << flush;
        exit( 1 );
      }
      _getKernel2 = &Kernel::_getKernel_invalid2;
      break;
    case TREE_FAST:
      // initialize node look-up data
      for ( int i = 0; i < example.l; i++ ) {
        if ( example.x[i]->extra && example.x[i]->extra->getTree() ) {
          example.x[i]->extra->getTree()->initNodeLookup();   
        } else {
          cerr << "There is no tree part\n"; exit(1);
        }
      }
      if ( node_restriction_type == NO_RESTRICT ) {
        if ( geometric_type == NO_GEOMETRIC ) {
          _getKernel  = &Kernel::_getKernel_unknown;
          _getKernel3 = &Kernel::_getKernel_tree_common3;
        } else if ( geometric_type == GEOMETRIC_1 ) {
          _getKernel  = &Kernel::_getKernel_unknown;
          _getKernel3 = &Kernel::_getKernel_default3;
          //_getKernel = &Kernel::_getKernel_tree_geo1;
        } else {
          cerr << "Unsupported geometric type: " 
                    << geometric_type << "\n" << flush;
          exit( 1 );
        }
      } else if ( node_restriction_type == RESTRICT_1 ) {
        if ( geometric_type == NO_GEOMETRIC ) {
          _getKernel  = &Kernel::_getKernel_tree_restrict1_fast;
          _getKernel3 = &Kernel::_getKernel_tree_common3;
        } else if ( geometric_type == GEOMETRIC_1 ) {
          _getKernel  = &Kernel::_getKernel_unknown;
          _getKernel3 = &Kernel::_getKernel_tree_common3;
          //_getKernel = &Kernel::_getKernel_tree_restrict1_geo1;
        } else {
          cerr << "Unsupported geometric type: " 
                    << geometric_type << "\n" << flush;
          exit( 1 );
        }
      } else {
        cerr << "Unsupported restriction type: " 
                  << node_restriction_type << "\n" << flush;
        exit( 1 );
      }
      _getKernel2 = &Kernel::_getKernel_invalid2;
      break;
    case TREE_ELASTIC: // J.K.
      if ( node_restriction_type == NO_RESTRICT ) {
        if ( geometric_type == NO_GEOMETRIC ) {
          _getKernel  = &Kernel::_getKernel_tree_elastic;
          _getKernel3 = &Kernel::_getKernel_tree_common3;
        } else if ( geometric_type == GEOMETRIC_1 ) {
          _getKernel = &Kernel::_getKernel_unknown;
          _getKernel3 = &Kernel::_getKernel_tree_common3;
          //_getKernel = &Kernel::_getKernel_tree_elastic_geo1;
        } else {
          cerr << "Unsupported geometric type: " 
                    << geometric_type << "\n" << flush;
          exit( 1 );
        }
      } else if ( node_restriction_type == RESTRICT_1 ) {
        if ( geometric_type == NO_GEOMETRIC ) {
          _getKernel = &Kernel::_getKernel_tree_elastic_restrict1;
          _getKernel3 = &Kernel::_getKernel_tree_common3;
        } else if ( geometric_type == GEOMETRIC_1 ) {
          _getKernel = &Kernel::_getKernel_unknown;
          _getKernel3 = &Kernel::_getKernel_tree_common3;
          //_getKernel = &Kernel::_getKernel_tree_elastic_restrict1_geo1;
        } else {
          cerr << "Unsupported geometric type: " 
                    << geometric_type << "\n" << flush;
          exit( 1 );
        }
      } else {
         cerr << "Unsupported restriction type: " 
                  << node_restriction_type << "\n" << flush;
        exit( 1 );
      }
      _getKernel2 = &Kernel::_getKernel_invalid2;
      break;
    case TREE_COLLINS:
      if ( node_restriction_type == NO_RESTRICT ) {
        if ( geometric_type == NO_GEOMETRIC ) {
          _getKernel = &Kernel::_getKernel_unknown;
          _getKernel3 = &Kernel::_getKernel_tree_common3;
        } else if ( geometric_type == GEOMETRIC_1 ) {
          _getKernel = &Kernel::_getKernel_unknown;
          _getKernel3 = &Kernel::_getKernel_tree_common3;
        } else {
          cerr << "Unsupported geometric type: " 
                    << geometric_type << "\n" << flush;
          exit( 1 );
        }
      } else if ( node_restriction_type == RESTRICT_1 ) {
        if ( geometric_type == NO_GEOMETRIC ) {
          _getKernel = &Kernel::_getKernel_tree_collins_restrict1;
          _getKernel3 = &Kernel::_getKernel_tree_common3;
        } else if ( geometric_type == GEOMETRIC_1 ) {
          _getKernel = &Kernel::_getKernel_unknown;
          _getKernel3 = &Kernel::_getKernel_tree_common3;
        } else {
          cerr << "Unsupported geometric type: " 
                    << geometric_type << "\n" << flush;
          exit( 1 );
        }
      } else {
        cerr << "Unsupported restriction type: " 
                  << node_restriction_type << "\n" << flush;
        exit( 1 );
      }
      _getKernel2 = &Kernel::_getKernel_invalid2;
      break;
    case TREE_COLLINS_FAST:
      // initialize node look-up data
      for ( int i = 0; i < example.l; i++ ) {
        if ( example.x[i]->extra && example.x[i]->extra->getTree() ) {
          example.x[i]->extra->getTree()->initNodeLookup();   
        } else {
          cerr << "There is no tree part\n"; exit(1);
        }
      }
      if ( node_restriction_type == NO_RESTRICT ) {
        if ( geometric_type == NO_GEOMETRIC ) {
          _getKernel = &Kernel::_getKernel_unknown;
          _getKernel3 = &Kernel::_getKernel_tree_common3;
        } else if ( geometric_type == GEOMETRIC_1 ) {
          _getKernel = &Kernel::_getKernel_unknown;
          _getKernel3 = &Kernel::_getKernel_tree_common3;
        } else {
          cerr << "Unsupported geometric type: " 
                    << geometric_type << "\n" << flush;
          exit( 1 );
        }
      } else if ( node_restriction_type == RESTRICT_1 ) {
        if ( geometric_type == NO_GEOMETRIC ) {
          _getKernel = &Kernel::_getKernel_tree_collins_restrict1_fast;
          _getKernel3 = &Kernel::_getKernel_tree_common3;
        } else if ( geometric_type == GEOMETRIC_1 ) {
          _getKernel = &Kernel::_getKernel_unknown;
          _getKernel3 = &Kernel::_getKernel_tree_common3;
        } else {
          cerr << "Unsupported geometric type: " 
                    << geometric_type << "\n" << flush;
          exit( 1 );
        }
      } else {
        cerr << "Unsupported restriction type: " 
                  << node_restriction_type << "\n" << flush;
        exit( 1 );
      }
      _getKernel2 = &Kernel::_getKernel_invalid2;
      break;
    case TREE_COLLINS_PLUS: // J.K.
      if ( node_restriction_type == NO_RESTRICT ) {
        if ( geometric_type == NO_GEOMETRIC ) {
          _getKernel = &Kernel::_getKernel_unknown;
          _getKernel3 = &Kernel::_getKernel_tree_common3;
        } else if ( geometric_type == GEOMETRIC_1 ) {
          _getKernel = &Kernel::_getKernel_unknown;
          _getKernel3 = &Kernel::_getKernel_tree_common3;
          //_getKernel = &Kernel::_getKernel_tree_elastic_geo1;
        } else {
          cerr << "Unsupported geometric type: " 
                    << geometric_type << "\n" << flush;
          exit( 1 );
        }
      } else if ( node_restriction_type == RESTRICT_1 ) {
        if ( geometric_type == NO_GEOMETRIC ) {
          _getKernel = &Kernel::_getKernel_tree_collins_plus_restrict1;
          _getKernel3 = &Kernel::_getKernel_tree_common3;
        } else if ( geometric_type == GEOMETRIC_1 ) {
          _getKernel = &Kernel::_getKernel_unknown;
          _getKernel3 = &Kernel::_getKernel_tree_common3;
          //_getKernel = &Kernel::_getKernel_tree_elastic_restrict1_geo1;
        } else {
          cerr << "Unsupported geometric type: " 
                    << geometric_type << "\n" << flush;
          exit( 1 );
        }
      } else {
         cerr << "Unsupported restriction type: " 
                  << node_restriction_type << "\n" << flush;
        exit( 1 );
      }
      _getKernel2 = &Kernel::_getKernel_invalid2;
      break;
    case TREE_COLLINS_P:
      // initialize Collins production tree
      for ( int i = 0; i < example.l; i++ ) {
        if ( example.x[i]->extra 
             && example.x[i]->extra->getTree() ) {
          if (!example.x[i]->extra->getTreeCollinsP() ) {
            //std::cerr << "make collp_tree\n" << std::flush;
            LOTree* collp_tree 
              = example.x[i]->extra->getTree()->makeCollinsPLOTree( production_map );
            example.x[i]->extra->setTreeCollinsP( collp_tree );
          }
        } else {
          cerr << "There is no tree part\n"; exit(1);
        }
      }
      if ( node_restriction_type == NO_RESTRICT ) {
        if ( geometric_type == NO_GEOMETRIC ) {
          _getKernel = &Kernel::_getKernel_unknown;
          _getKernel3 = &Kernel::_getKernel_tree_common3;
        } else if ( geometric_type == GEOMETRIC_1 ) {
          _getKernel = &Kernel::_getKernel_unknown;
          _getKernel3 = &Kernel::_getKernel_tree_common3;
        } else {
          cerr << "Unsupported geometric type: " 
                    << geometric_type << "\n" << flush;
          exit( 1 );
        }
      } else if ( node_restriction_type == RESTRICT_1 ) {
        if ( geometric_type == NO_GEOMETRIC ) {
          _getKernel = &Kernel::_getKernel_tree_collins_p_restrict1;
          _getKernel3 = &Kernel::_getKernel_tree_common3;
        } else if ( geometric_type == GEOMETRIC_1 ) {
          _getKernel = &Kernel::_getKernel_unknown;
          _getKernel3 = &Kernel::_getKernel_tree_common3;
        } else {
          cerr << "Unsupported geometric type: " 
                    << geometric_type << "\n" << flush;
          exit( 1 );
        }
      } else {
        cerr << "Unsupported restriction type: " 
                  << node_restriction_type << "\n" << flush;
        exit( 1 );
      }
      _getKernel2 = &Kernel::_getKernel_invalid2;
      break;
    case TREE_COLLINS_P_FAST:
      // initialize Collins production tree and node look-up data
      for ( int i = 0; i < example.l; i++ ) {
        if ( example.x[i]->extra 
             && example.x[i]->extra->getTree() ) {
          if (!example.x[i]->extra->getTreeCollinsP() ) {
            //std::cerr << "AAA\n" << flush;
            LOTree* collp_tree 
              = example.x[i]->extra->getTree()->makeCollinsPLOTree( production_map );
            example.x[i]->extra->setTreeCollinsP( collp_tree );
            //std::cerr << collp_tree->toString() << "\n" << std::flush;
          }
          example.x[i]->extra->getTreeCollinsP()->initNodeLookup();
        } else {
          cerr << "There is no tree part\n"; exit(1);
        }
      }
      if ( node_restriction_type == NO_RESTRICT ) {
        if ( geometric_type == NO_GEOMETRIC ) {
          _getKernel = &Kernel::_getKernel_unknown;
          _getKernel3 = &Kernel::_getKernel_tree_common3;
        } else if ( geometric_type == GEOMETRIC_1 ) {
          _getKernel = &Kernel::_getKernel_unknown;
          _getKernel3 = &Kernel::_getKernel_tree_common3;
        } else {
          cerr << "Unsupported geometric type: " 
                    << geometric_type << "\n" << flush;
          exit( 1 );
        }
      } else if ( node_restriction_type == RESTRICT_1 ) {
        if ( geometric_type == NO_GEOMETRIC ) {
          _getKernel = &Kernel::_getKernel_tree_collins_p_restrict1_fast;
          _getKernel3 = &Kernel::_getKernel_tree_common3;
        } else if ( geometric_type == GEOMETRIC_1 ) {
          _getKernel = &Kernel::_getKernel_unknown;
          _getKernel3 = &Kernel::_getKernel_tree_common3;
        } else {
          cerr << "Unsupported geometric type: " 
                    << geometric_type << "\n" << flush;
          exit( 1 );
        }
      } else {
        cerr << "Unsupported restriction type: " 
                  << node_restriction_type << "\n" << flush;
        exit( 1 );
      }
      _getKernel2 = &Kernel::_getKernel_invalid2;
      break;
    default:
      fprintf(stderr,"Kernel::Kernel: Unknown kernel function\n");
    }

    cerr << "Calculate K(x,x) for all examples. l=" << l << "\n" << flush;
    double max_kk = (this->*_getKernel)( example.x[0], example.x[0] );
    for( int i = 0; i < l; i++ ) {
      double k = (this->*_getKernel)( example.x[i], example.x[i] );
      if ( k > max_kk ) {
        max_kk = k;
      }
      example.x[i]->extra->setSelfKernelSqrt( sqrt(k) );
    }
    cerr << "Maximum of k(x,x)=" << max_kk << "\n" << flush;

    if ( param.auto_param_s && param.param_s < 0.0 ) {
      // the second condition to ensure that initialization takes place only once
      cerr << "Calculate param_s automatically.\n" << flush;
      double max_k = 0.0;
      for( int i = 0; i < l; i++ ) {
        double d = example.x[i]->extra->getSelfKernelSqrt();
        double k = d * d;
        if ( k > max_k ) { max_k = k; }
      }
      if ( max_k == 0.0 ) { max_k = 1.0; }

      double d = 1.0/max_k;
      double new_param_s = d * param.param_s;
      cerr << "Set param_s to " << d << "*" << param.param_s 
                << "=" << new_param_s << flush;
      param_s       = new_param_s;
      param.param_s = new_param_s;
    }
  }

  ~Kernel() 
  {
    if ( tree_c ) {
      _free_matrix( tree_c, max_tree_size, max_tree_size );
    }
    if ( tree_s ) {
      _free_matrix( tree_s, max_tree_size, max_tree_size );
    }
    if ( tree_ca ) {
      _free_matrix( tree_ca, max_tree_size, max_tree_size );
    }
    if ( tree_cr ) {
      _free_matrix( tree_cr, max_tree_size, max_tree_size );
    }
    if ( tree_sr ) {
      _free_matrix( tree_sr, max_tree_size, max_tree_size );
    }
    if ( tree_car ) {
      _free_matrix( tree_car, max_tree_size, max_tree_size );
    }
  }

  // wrapper for getKernel
  inline 
  double getKernel(const feature_node *x1, const feature_node *x2) const
  {
    double k = 0.0;
    if ( self_normalize ) { // J.K
      //cerr << "AA\n" << flush;
      double x1_x1 = x1->extra->getSelfKernelSqrt();
      double x2_x2 = x2->extra->getSelfKernelSqrt();
      double x1_x2 = (this->*_getKernel)(x1, x2);
      if ( x1_x1 < 0.0 ) {
        x1_x1 = sqrt( (this->*_getKernel)(x1, x1) );
      }
      if ( x2_x2 < 0.0 ) {
        x2_x2 = sqrt( (this->*_getKernel)(x2, x2) );
      }
      //cerr << "x1_x1=" << x1_x1 << " x2_x2=" << x2_x2 << " x1_x2=" << x1_x2 << "\n" << flush;
      k = x1_x2 / (x1_x1 * x2_x2);
    } else {
      k = (this->*_getKernel)(x1, x2);
    }

    return k;
  }

  inline double getKernel(const double _x) const 
  {
    return (this->*_getKernel2)(_x);
  }
  
  inline void getKernel( const feature_node*   x,
                         feature_node** const  xs, // p to p to const fn
                         unsigned int          size,
                         double*               dest ) const
  {
    (this->*_getKernel3)( x, xs, size, dest );
    if ( self_normalize ) {
      double x1_x1 = x->extra->getSelfKernelSqrt();
      if ( x1_x1 < 0.0 ) {
        x1_x1 = sqrt( (this->*_getKernel)(x, x) );
      }
      for ( unsigned int k = 0; k < size; k++ ) {
        double x2_x2 = xs[k]->extra->getSelfKernelSqrt();
         if ( x2_x2 < 0.0 ) {
           x2_x2 = sqrt( (this->*_getKernel)(xs[k], xs[k]) );
         }
         dest[k] = dest[k] / ( x1_x1 * x2_x2 );
      }
    }
  }
 };
};
#endif
