/*
 This file is added by J. Kazama to the TinySVM package.
 Copyright (C) 2005. Jun'ichi Kazama <kazama@jaist.ac.jp>
 All rights reserved.

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Library General Public
 License as published by the Free Software Foundation; either
 version 2 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Library General Public License for more details.

 You should have received a copy of the GNU Library General Public
 License along with this library; if not, write to the
 Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 Boston, MA 02111-1307, USA.
*/

#ifndef _LOTREE_H
#define _LOTREE_H

// J.K.
// labeled ordered tree
// with geometric and node restriction information if necessary 

#include <vector>
#include <string>
#include <sstream>
#include <iostream>
#include <assert.h>
#include <map>

#include "../config.h"
//#include "darts.h"

namespace TinySVM {
  /**
   *
   */
  class Rectangle {
  public:
    double x,y,w,h;
    Rectangle( double _x, double _y, double _w, double _h ) {
      x = _x; y = _y; w = _w; h = _h;
    }
  };

  /**
   *
   */
  class LOTreeNode {
  public:
    static const int  one_key_len = 2*sizeof( int );
    
  protected:
    int          post_ordered_no;
#ifdef LABEL_INTEGER
    typedef int label_t;
#else
    typedef std::string label_t;
#endif
    
    label_t label;

    LOTreeNode*  parent;
    LOTreeNode*  first_child;
    LOTreeNode*  next_sibling;
    Rectangle*   geometry;
    int          mark; // -1 -- not marked, 0 ... -- marked with number
    int          num_children;
    LOTreeNode*  nearest_marked_node;

    /*
    typedef 
    Darts::DoubleArrayImpl<char, unsigned char, int, unsigned int> trie_t;
    */

  public:
    LOTreeNode( ) 
      : post_ordered_no( -1 ), parent( 0 ), 
        first_child( 0 ), next_sibling( 0 ), 
        geometry( 0 ), mark( -1 ), num_children( 0 ),
        nearest_marked_node( 0 )
    {  
      
    };
    
    ~LOTreeNode() 
    {
      /*
      LOTreeNode* ch = getFirstChild();
      while ( ch ) {
        LOTreeNode* ns = ch->getNextSibling();
        delete ch;
        ch = ns;
      }
      */
      if ( geometry ) { delete geometry; }
    }
#ifdef LABEL_INTEGER
    const int getLabel() const { return label; }
    void setLabel( const int _label ) { label = _label; }
#else
    const std::string& getLabel() const { return label; }
    void setLabel( const std::string& _label ) { label = _label; }
#endif
    

    LOTreeNode* getParent() { return parent; }
    LOTreeNode* getFirstChild()  { return first_child; }
    LOTreeNode* getNextSibling() { return next_sibling; }

    const LOTreeNode* getParent() const { return parent; }
    const LOTreeNode* getFirstChild()  const { return first_child; }
    const LOTreeNode* getNextSibling() const { return next_sibling; }

    void appendChild( LOTreeNode* _child ) {
      if ( !first_child ) {
        first_child = _child;
      } else {
        LOTreeNode* last_ch = first_child;
        while( last_ch->getNextSibling() ) {
          last_ch = last_ch->getNextSibling();
        }
        last_ch->setNextSibling( _child );
      }
      _child->parent = this;
    }

    void setNextSibling( LOTreeNode* _next ) {
      next_sibling = _next;
    }

    void setGeometry( Rectangle* _rect ) {
      if ( geometry ) { delete geometry; }
      geometry = _rect;
    }

    void setMark( int _m ) { mark = _m; }
    bool isMarked() const { return (mark > -1); }
    int  getMark() const { return mark; }

    int calcNumNodesRec() const {
      int n = 1;
      const LOTreeNode* ch = getFirstChild();
      while ( ch ) {
        n += ch->calcNumNodesRec();
        ch = ch->getNextSibling();
      }
      return n; 
    }

    int numChildren() const {
      return num_children;
    }

    int calcNumChildren() const {
      int n = 0;
      const LOTreeNode* ch = getFirstChild();
      while ( ch ) {
        n++;
        ch = ch->getNextSibling();
      }
      return n; 
    }

    void initNumChildrenRec() {
      num_children = calcNumChildren();
      LOTreeNode* ch = getFirstChild();
      while( ch ) {
        ch->initNumChildrenRec();
        ch = ch->getNextSibling();
      }
    }

    int calcMaxNumChildrenRec() const {
      int ret = 0;
      const LOTreeNode* ch = getFirstChild();
      while ( ch ) {
        int m = ch->calcMaxNumChildrenRec();
        if ( m > ret ) { ret = m; }
        ch = ch->getNextSibling();
      }
      return ret;
    }

    bool isLeaf() const {
      return ( getFirstChild() == 0 );
    }


    bool containAtLeastOneMarkRec() const {
      if ( isMarked() ) { return true; }
      const LOTreeNode* ch = getFirstChild();
      while ( ch ) {
        if ( ch->containAtLeastOneMarkRec() ) { 
          return true;
        } else {
          ch = ch->getNextSibling();
        }
      }
      return false;
    }

    std::string toString( bool rec = true, bool out_mark = true, bool out_geo = true ) const {
      std::ostringstream oss;
      writeToStream( oss, rec, out_mark, out_geo );
      return oss.str();
    }

    std::string toGraphvizString( bool rec = true, bool out_mark = true, bool out_geo = true ) const {
      std::ostringstream oss;
      writeToStreamForGraphviz( oss, rec, out_mark, out_geo );
      return oss.str();
    }

    void _get_all_subtree_combination(const std::string& prefix, 
                                      const std::vector< std::vector<std::string> >& ch_v, unsigned int ch_v_i,
                                      std::vector< std::string >& to ) const {
      if ( ch_v_i == ch_v.size() ) {
        to.push_back( prefix + ')');
      } else {
        const std::vector< std::string >& v = ch_v[ ch_v_i ];
        _get_all_subtree_combination( prefix, ch_v, ch_v_i+1, to );
        for ( unsigned int i = 0; i < v.size(); i++ ) {
          _get_all_subtree_combination( prefix + ' ' + v[i], ch_v, ch_v_i+1, to );
        }
      }
    }
    
    void getAllSubtreeStringRec_sub( std::vector< std::string >& to, bool out_mark = true, bool out_geo = true ) const {
      std::string this_s = this->toString( false, out_mark, out_geo );
      std::vector< std::vector< std::string > > ch_v;
      
      const LOTreeNode* ch = getFirstChild();
      while ( ch ) {
        std::vector< std::string > v( 0 );
        ch->getAllSubtreeStringRec_sub( v, out_mark, out_geo );
        ch_v.push_back( v );
        ch = ch->getNextSibling();
      }
      
      _get_all_subtree_combination( this_s.substr( 0, this_s.length()-1 ), ch_v, 0, to );
    }
    
    void getAllSubtreeStringRec( std::vector< std::string >& to, bool out_mark = true, bool out_geo = true ) const {
      this->getAllSubtreeStringRec_sub( to, out_mark, out_geo ); // starting with this node
      
      const LOTreeNode* ch = getFirstChild();
      while ( ch ) {
        ch->getAllSubtreeStringRec( to, out_mark, out_geo );
        ch = ch->getNextSibling();
      }
    }

    void getAllElasticSubtreeStringRec_sub( std::vector< std::string >& to, bool out_mark = true, bool out_geo = true ) const {
      std::string this_s = this->toString( false, out_mark, out_geo );
      std::vector< std::vector< std::string > > ch_v;
      
      const LOTreeNode* ch = getFirstChild();
      while ( ch ) {
        std::vector< std::string > v( 0 );
        ch->getAllElasticSubtreeStringRec( v, out_mark, out_geo );
        ch_v.push_back( v );
        ch = ch->getNextSibling();
      }
      
      _get_all_subtree_combination( this_s.substr( 0, this_s.length()-1 ), ch_v, 0, to );
    }

    void getAllElasticSubtreeStringRec( std::vector< std::string >& to, bool out_mark = true, bool out_geo = true) const {
      this->getAllElasticSubtreeStringRec_sub( to, out_mark, out_geo );
      const LOTreeNode* ch = getFirstChild();
      while ( ch ) {
        ch->getAllElasticSubtreeStringRec( to, out_mark, out_geo );
        ch = ch->getNextSibling();
      }
    }

    void writeToStream( std::ostream& _os, bool rec = true, bool out_mark = true, bool out_geo = true ) const {
      _os << '(';
      if ( out_mark && isMarked() ) {
        if ( mark == 0 ) {
          _os << '*';
        }
        else {
          _os << '*' << mark;
        }
      }
      _os << '[';
#ifdef LABEL_INTEGER 
      _os << label;
#else
      _os << escape_char( label );
#endif
      _os << ']';

      if ( out_geo && geometry ) {
        _os << '<';
        _os << geometry->x << ' ';
        _os << geometry->y << ' ';
        _os << geometry->w << ' ';
        _os << geometry->h;
        _os << '>';
      }
      if ( rec ) {
        const LOTreeNode* ch = getFirstChild();
        while ( ch ) {
          _os << ' ';
          ch->writeToStream( _os, rec, out_mark, out_geo );
          ch = ch->getNextSibling();
        }
      }
      _os << ')';
    }

    void writeToStreamForGraphviz( std::ostream& _os, 
                                   bool rec = true, 
                                   bool out_mark = true, 
                                   bool out_geo = true ) const {
      // output node definition
      _os << "node" << post_ordered_no << " [label=\"";
#ifdef LABEL_INTEGER 
      _os << label;
#else
      _os << escape_char( label );
#endif
      if ( out_mark && isMarked() ) {
        if ( mark == 0 ) {
          _os << "(*)\"";
        }
        else {
          _os << "(*" << mark << ")\"";
        }
        _os << ", color=lightblue2, stile=filled]";
      } else {
        _os << "\"]";
      }
      _os << ";\n";
        
      if ( rec ) {
        const LOTreeNode* ch = getFirstChild();
        while ( ch ) {
          // output edge
          _os << "node" << post_ordered_no << " -> ";
          _os << "node" << ch->getNoPostOrdered() << ";\n";
          ch->writeToStreamForGraphviz( _os, rec, out_mark, out_geo );
          ch = ch->getNextSibling();
        }
      }
    }

    
    void initPostOrderedRec( std::vector<LOTreeNode*>& _to, 
                             int* _no ) {
      LOTreeNode* ch = getFirstChild();
      while ( ch ) {
        ch->initPostOrderedRec( _to, _no );
        ch = ch->getNextSibling();
      }

      _to[ (*_no) ] = this;
      this->setNoPostOrdered( *_no );
      (*_no)++;
    }

    void setNoPostOrdered( int _n ) {
      post_ordered_no = _n;
    }
    int getNoPostOrdered() const { return post_ordered_no; }

    // set the nearest next marked node 
    // that we meet after this node 
    // in the preorder traversal of the depth-first search
    LOTreeNode* initNearestMarkedNodeRec( LOTreeNode* prev = 0 ) {
      LOTreeNode* ch = getFirstChild();
      std::vector< LOTreeNode* > ch_v;
      while ( ch ) {
        ch_v.push_back( ch );
        ch = ch->getNextSibling();
      }

      LOTreeNode* nearest = prev;
      for( int i = (int)ch_v.size()-1; i >= 0; i-- ) {
        nearest = ch_v[i]->initNearestMarkedNodeRec( nearest ); 
      }

      this->nearest_marked_node = nearest;
      if ( this->isMarked() ) {
        return this;
      } else {
        return nearest;
      }
    }

    const LOTreeNode* getNearestMarkedNode() const {
      return nearest_marked_node;
    }

    typedef 
    std::pair< std::pair< label_t, int>, 
               std::pair< label_t, int> >
    edge_t;
    
    typedef 
    std::map< edge_t, int >
    bigram_pat_t;

    void getEdgePattern( bigram_pat_t& to ) const {
            
      const LOTreeNode* ch = getFirstChild();
      while ( ch ) {
        edge_t edge( std::make_pair( getLabel(), getMark() ),
                     std::make_pair( ch->getLabel(), ch->getMark() ) );
        to[edge]++;
        ch->getEdgePattern( to );
        ch = ch->getNextSibling();
      }
    }

    // for Trie construction
#ifdef LABEL_INTEGER
    void getPatternVec( std::vector< int >& to ) const {
      to.push_back( getLabel() ); 
      to.push_back( getMark()  ); 

      const LOTreeNode* ch = getFirstChild();
      while ( ch ) { 
        ch->getPatternVec( to );
        to.push_back( -1 ); 
        to.push_back( -1 ); 
        ch = ch->getNextSibling();
      }
    }
#else
    void getPatternVec( std::vector< unsigned int >& to ) const {
      std::cerr << "getPatternVec() is supported only with LABEL_INTEGER.\n" << flsuh;
      exit(1);
    }
#endif

  public:
    static std::string escape_char( const std::string& _label ) {
      std::ostringstream oss;
      for ( unsigned int i = 0; i < _label.length(); i++ ) {
        char c = _label[ i ];
        if ( c == '(' || c == ')' || c == '[' || c == ']' 
             || c == '#' || c == '\\' ) {
          oss << '\\' << c;
        } else {
          oss << c;
        }
      }
      return oss.str();
    }

#ifdef LABEL_INTEGER
    void
    get_trie_search_key( char* key, int label, int mark ) const 
    {
      int len = sizeof( int );
      int index = 0;
      const char* a = reinterpret_cast< const char* >( &label );
      for ( int k = 0; k < len ; k++ ) {
        key[index++] = a[k];
      }
      a = reinterpret_cast< const char* >( &mark );
      for ( int k = 0; k < len ; k++ ) {
        key[index++] = a[k];
      }
    }

    /*
    void 
    get_all_feature_sub( trie_t* trie, size_t _node_pos, int depth,
                         std::map< unsigned int, unsigned int>& to,
                         char* tmp_key ) const
    {
      const LOTreeNode* pnode    = this;
      size_t            node_pos = _node_pos; // pos in trie
      size_t np, kp;
      int    res;
      for (int d = -1 ; d < depth && pnode != 0; ++d) {
        const LOTreeNode* start = 0;
        if (d == -1) {
          start = pnode->getFirstChild();
        } else {
          np    = node_pos;
          get_trie_search_key( tmp_key, -1, -1 );
          kp = 0;
          res = trie->traverse( tmp_key, np, kp, one_key_len );
          if ( res != -2 ) {
            start    = pnode->getNextSibling();
            node_pos = np;
          } else {
            start = 0;
          }
        }
        int new_depth = depth - d;
        for ( const LOTreeNode* l = start; l != 0; l = l->getNextSibling() ) {
          np = node_pos; 
          get_trie_search_key( tmp_key, l->getLabel(), l->getMark() );
          kp = 0;
          res = trie->traverse( tmp_key, np, kp, one_key_len );
          if ( res >= 0 ) { // there is feature here
            to[ res ]++;
          }
          if ( res != -2 ) { // should search further
            l->get_all_feature_sub( trie, np, new_depth, to, tmp_key ); 
          }
        }
        if (d != -1) pnode = pnode->getParent();
      }
    }
    */

    /*
    void getAllFeature( trie_t* trie,
                        std::map< unsigned int, unsigned int>& to,
                        char* tmp_key ) const
    {

      // size one 
      size_t node_pos = 0;
      size_t kp = 0;
      get_trie_search_key( tmp_key, getLabel(), getMark() );
      int res = trie->traverse( tmp_key, node_pos, kp, one_key_len );
      if ( res >= 0 ) { // there is feature here
        to[ res ]++;
      }
      if ( res != -2 ) { // should search further
        this->get_all_feature_sub( trie, node_pos, 0, to, tmp_key );
      }

      const LOTreeNode* ch = getFirstChild();
      while ( ch ) { 
        ch->getAllFeature( trie, to, tmp_key );
        ch = ch->getNextSibling();
      }
    }
    */

    bool is_right_most_node( const LOTreeNode* node,
                             int   depth ) const {
      if ( depth == 0 || node == 0 ) { return true ; }
      else if ( node->getNextSibling() ) { return false; }
      else { return is_right_most_node( node->getParent(), depth-1 ); }
    }
    
    /*
    void 
    get_all_feature_collins_sub( trie_t* trie, 
                                 size_t _node_pos, int depth,
                                 std::map< unsigned int, unsigned int>& to,
                                 char* tmp_key ) const
    {
      const LOTreeNode* pnode    = this;
      size_t            node_pos = _node_pos; // pos in trie
      size_t np, kp;
      int    res;
      for (int d = -1 ; d < depth && pnode != 0; ++d) {
        const LOTreeNode* start = 0;
        if (d == -1) {
          start = pnode->getFirstChild();
        } else {
          np    = node_pos;
          get_trie_search_key( tmp_key, -1, -1 );
          kp = 0;
          res = trie->traverse( tmp_key, np, kp, one_key_len );
          if ( res != -2 ) {
            start    = pnode->getNextSibling();
            node_pos = np;
          } else {
            start = 0;
          }
        }
        int new_depth = depth - d;
        if ( start ) {
          const LOTreeNode* l = start;
          np = node_pos; 
          get_trie_search_key( tmp_key, l->getLabel(), l->getMark() );
          kp = 0;
          res = trie->traverse( tmp_key, np, kp, one_key_len );
          if ( res >= 0 && is_right_most_node( l, new_depth ) ) { // there is feature here
            to[ res ]++;
          }
          if ( res != -2 ) { // should search further
            l->get_all_feature_collins_sub( trie, np, new_depth, to, tmp_key ); 
          }
        }
        if ( d != -1 ) {
          if ( !pnode->getNextSibling() ) {
            pnode = pnode->getParent();
          } else {
            pnode = 0;
          }
        }
      }
    }
    
    void getAllFeatureCollins( trie_t* trie, 
                               std::map< unsigned int, unsigned int>& to,
                               char* tmp_key ) const
    {

      size_t node_pos = 0;
      size_t kp = 0;
      get_trie_search_key( tmp_key, getLabel(), getMark() );
      int res = trie->traverse( tmp_key, node_pos, kp, one_key_len );
      if ( res >= 0 ) { // there is feature here
        std::cerr << "this should not happen!\n" << std::flush;
        exit(1);
      }
      if ( res != -2 ) { // should search further
        this->get_all_feature_collins_sub( trie, node_pos, 0, to, tmp_key );
      }


      const LOTreeNode* ch = getFirstChild();
      while ( ch ) { 
        ch->getAllFeatureCollins( trie, to, tmp_key );
        ch = ch->getNextSibling();
      }
    }

    */

    typedef 
    std::map< std::vector< std::pair< int, int> >, int >  
    production_map_t;
    
    void makeCollinsPProductionMap( production_map_t& to ) const {
      if ( !isLeaf() ) {
        std::vector< std::pair<int, int> > prod_v;
        
        bool marked = isMarked();
        prod_v.push_back( std::make_pair( getLabel(), getMark() ) );
        
        const LOTreeNode* ch = getFirstChild();
        while ( ch ) { 
          prod_v.push_back( std::make_pair( ch->getLabel(), ch->getMark() ) );
          marked |= ch->isMarked();
          ch = ch->getNextSibling();
        }
        if ( to.count( prod_v ) == 0 ) {
          to[ prod_v ] = to.size(); // register new production
        }

        ch = getFirstChild();
        while ( ch ) { 
          ch->makeCollinsPProductionMap( to );
          ch = ch->getNextSibling();
        }
      }
    }


    LOTreeNode* makeCollinsPLOTree( production_map_t& prod_map ) const
    {
      if ( isLeaf() ) {
        return 0;
      } else {
        std::vector< std::pair<int,int> > prod_v;
        bool marked = isMarked();
        prod_v.push_back( std::make_pair( getLabel(), getMark() ) );
        
        const LOTreeNode* ch = getFirstChild();
        while ( ch ) { 
          prod_v.push_back( std::make_pair( ch->getLabel(), ch->getMark() ) );
          marked |= ch->isMarked();
          ch = ch->getNextSibling();
        }

        int prod_id = prod_map.size(); // unknown production;
        int mark    = marked ? 0 : -1; // 
        production_map_t::iterator pi = prod_map.find( prod_v );
        if ( pi != prod_map.end() ) {
          prod_id = pi->second;
        } else {
          prod_map[ prod_v ] = prod_id;
        }
        LOTreeNode* node = new LOTreeNode();
        node->setLabel( prod_id );
        node->setMark( mark );
        ch = getFirstChild();
        while ( ch ) { 
          LOTreeNode* n = ch->makeCollinsPLOTree( prod_map );
          if ( n ) {
            node->appendChild( n );
          }
          ch = ch->getNextSibling();
        }
        return node;
      }
    }
#endif

  };

  /**
   * 
   */
  class LOTree {
  private:
    LOTreeNode*              root;
    int                      num_nodes;
    int                      max_num_children;
    // node_no => node (node_no is assigned in the post-ordered manner)
    std::vector<LOTreeNode*> post_ordered;  
    // label => node_no
#ifdef LABEL_INTEGER
    typedef int label_t;
#else
    typedef std::string label_t;
#endif

    std::map< label_t, std::vector< int > > label2nodes;
    std::vector< int >                      dummy_nodes;

    bool node_lookup_initialized;
    
    
    /*
    typedef 
    Darts::DoubleArrayImpl<char, unsigned char, int, unsigned int> trie_t;
    */

  public:
    LOTree() : 
      root( 0 ), num_nodes( 0 ), post_ordered( 0 ), dummy_nodes( 0 ),
      node_lookup_initialized( false ) 
    { 
    }
   
    ~LOTree() { 
      //if ( root ) { delete root; }
      for(unsigned int i = 0; i < post_ordered.size(); i++ ) {
        delete post_ordered[i];
      }
    }

    void initInfo() {
       num_nodes = root->calcNumNodesRec();
       max_num_children = root->calcMaxNumChildrenRec();
       post_ordered = std::vector< LOTreeNode* >( num_nodes );
       int no = 0;
       root->initPostOrderedRec( post_ordered, &no );
       root->initNumChildrenRec();
       root->initNearestMarkedNodeRec();
       //initNodeLookup();
    }

    LOTree( LOTreeNode* _root ) :
      root( 0 ), num_nodes( 0 ), post_ordered( 0 ), dummy_nodes( 0 ),
      node_lookup_initialized( false ) 
    {
      root = _root;
      initInfo();
    }

    LOTree( const std::string& _str ) :
      root( 0 ), num_nodes( 0 ), post_ordered( 0 ), dummy_nodes( 0 ),
      node_lookup_initialized( false ) 
    {
      // std::cerr << "line=" << _str << "\n" << std::flush;
      enum parse_state {INIT, LABEL, LABEL_CONTENT, LABEL_ESCAPE, GEO, GEO_CONTENT};
      
      std::string label_buf( 32, ' ');
      label_buf.resize( 0 );
      std::string geo_buf( 32, ' ');
      geo_buf.resize( 0 );

      LOTreeNode* parent  = 0;
      LOTreeNode* current = 0;

      int state = INIT; // start with the state reading the label of the root node
      for ( unsigned int i = 0; i < _str.length(); i++ ) {
        char c = _str[ i ];
        switch( state ) {
        case INIT:
          {
            if ( c == '(' ) {
              parent  = current;
              current = new LOTreeNode();
              if ( parent ) {
                parent->appendChild( current );
              } 
              state = LABEL;
            } else if ( c == ')' ) {
              if ( !current ) {
                std::cerr << "LOTree format error?. [too many ')']\n" << std::cerr;
                exit(1);
              }
              if ( parent == 0 ) {
                root = current;
              }
              current = parent;
              if ( current ) {
                parent  = current->getParent();
              }
            }
          break;
          }
        case LABEL:
          {
            if ( c == '[' ) {
              label_buf.clear();
              state = LABEL_CONTENT; 
            } else if ( c == '*' ) {
              unsigned int k = i+1;
              while ( k < _str.length() && _str[k] != '[') {
                k++;
              }
              int mark;
              if ( k == i+1 ) { // mark is * (then mark = 0)
                mark = 0;
              } else {
                mark = atoi( _str.substr( i+1, k - i - 1 ).c_str() );
              }
              current->setMark( mark );
            }
            break;
          }
        case LABEL_CONTENT:
          {
            if ( c == ']' ) { 
#ifdef LABEL_INTEGER
              char* endptr;
              int label = (int)strtol( label_buf.c_str(), &endptr, 10 );
              if ( *endptr == '\0' ) { // valid integer
                current->setLabel( label );
              } else {
                std::cerr << "Invalid label integer\n" << std::flush;
                exit(1);
              }
#else
              current->setLabel( label_buf );
#endif 
              label_buf.clear(); 
              if ( i+1 < _str.length() && _str[ i + 1 ] == '<') {
                state = GEO;
              } else {
                state = INIT;
              }
            } else if ( c == '\\' ) { 
              state = LABEL_ESCAPE; 
            } else { label_buf.push_back( c ); }
            break;
          }
        case LABEL_ESCAPE:
          {
            label_buf.push_back( c );
            state = LABEL_CONTENT;
            break;
          }
        case GEO:
          {
            if ( c == '<' ) {
              geo_buf.clear();
              state = GEO_CONTENT;
            } else {
              state = INIT;
            }
            break;
          }
        case GEO_CONTENT:
          {
            if ( c == '>' ) {
              std::istringstream iss( geo_buf );
              double x,y,w,h;
              iss >> x >> y >> w >> h;
              current->setGeometry( new Rectangle( x, y, w, h) );
              geo_buf.clear();
              state = INIT;
            } else {
              geo_buf.push_back( c );
            }
            break;
          }
        }
      }
      
      if ( parent != 0 || label_buf.size() > 0 || root == 0 ) {     
        std::cerr << "LOTree format error?\n" << std::flush;
        exit(1);
      } else {
        initInfo();
      }
    }

    
    LOTreeNode* getRootNode() {
      return root;
    }

    int numNodes() const { return num_nodes; }
    int maxNumChildren() const { return max_num_children; }

    const LOTreeNode* getNodePostOrdered( int _no ) const {
      //assert( _no >= 0 && _no < (int)post_ordered.size() );
      return post_ordered[ _no ];
    } 
    LOTreeNode* getNodePostOrdered( int _no ) {
      //assert( _no >= 0 && _no < (int)post_ordered.size() );
      return post_ordered[ _no ];
    } 

    std::string toString() const {
      if ( root ) { return root->toString(); }
    }

    std::string toGraphvizString( std::string graph_name ) const {
      if ( root ) { 
        return ("digraph " + graph_name + " {\n" + root->toGraphvizString() + "}");
      } else {
        return ("digraph " + graph_name + " {\n}");
      }
    }
    
    void getAllSubtreeStringRec( std::vector< std::string >& to, bool out_mark = true, bool out_geo = true ) const {
      if ( root ) { root->getAllSubtreeStringRec( to, out_mark, out_geo ); }
    }
    
    void getAllElasticSubtreeStringRec( std::vector< std::string >& to, bool out_mark = true, bool out_geo = true ) const {
      if ( root ) { root->getAllElasticSubtreeStringRec( to, out_mark, out_geo ); }
    }

    bool containAtLeastOneMarkRec() const {
      if ( root ) { return root->containAtLeastOneMarkRec(); }
    }

    void initNodeLookup() {
      if (!node_lookup_initialized) {
        for ( int i = 0; i < post_ordered.size(); i++ ) {
          label2nodes[ post_ordered[i]->getLabel() ].push_back( i );
        }
        node_lookup_initialized = true;
      }
    }

    const std::vector< int >& getSameLabelNodes( const label_t label ) const {
      std::map< label_t, std::vector< int > >::const_iterator itr 
        = label2nodes.find( label );
      if ( itr != label2nodes.end() ) {
        return itr->second;
      } else {
        return dummy_nodes;
      }
    }
    
    typedef 
    std::pair< std::pair< label_t, int>, 
               std::pair< label_t, int> >
    edge_t;

    typedef 
    std::map< edge_t, int >
    bigram_pat_t;

    void getEdgePattern( bigram_pat_t& to ) const {
      to.clear();
      if ( root ) {
        root->getEdgePattern( to );
      }
    }

    double getEdgePatternSimilarity( const LOTree* t2 ) const {
      bigram_pat_t p1;
      bigram_pat_t p2;
      getEdgePattern( p1 );
      t2->getEdgePattern( p2 );
      double sum = 0.0;
      for ( bigram_pat_t::iterator itr 
              = p1.begin(); itr != p1.end(); ++itr ) {
        const edge_t& edge = itr->first;
        bigram_pat_t::iterator i2 = p2.find( edge );
        if ( i2 != p2.end() ) {
          sum += itr->second * i2->second;
        }
      }
      return sum;
    }

#ifdef LABEL_INTEGER
    void getPatternVec( std::vector< int >& to ) const {
      to.resize( 0 );
      if ( root ) {
        root->getPatternVec( to );
        // remove last contiguous [-1,-1]
        while( to.size() >= 2 
               &&
               to[ to.size() - 1 ] == -1
               &&
               to[ to.size() - 2 ] == -1 ) {
          to.pop_back();
          to.pop_back();
        }
      }
    }

    std::string getPatternString() const {
      std::vector< int > pat_v;
      getPatternVec( pat_v );
      std::ostringstream oss;
      int len = sizeof( int );
      for( unsigned int i = 0; i < pat_v.size(); i++ ) {
        int   n = pat_v[i];
        const char* a = reinterpret_cast< const char* >( &n );
        for ( int k = 0; k < len; k++ ) {
          oss << a[k];
        }
      }
      /*
      std::string ret = oss.str();
      if ( ret.length() != pat_v.size() * len ) {
        std::cerr << "Error in getPatternString()\n" << std::flush;
      }
      */
      return oss.str();
    }
#else
    void getPatternVec( std::vector< unsigned int >& to ) const {
      std::err << "getPatternVec(..) is supported only with LABEL_INTEGER\n" << flush;
      exit(1);
    }
#endif

    /*
    // search all the subtree features
    void getAllFeature( trie_t* trie,
                        std::map< unsigned int, unsigned int>& to ) const
    {
      if ( root ) {
        char tmp_key[ LOTreeNode::one_key_len ];
        root->getAllFeature( trie, to, tmp_key );
      }
    }
    */

    /*
    // search all the Collins-type subtree features
    void getAllFeatureCollins( trie_t* trie,
                        std::map< unsigned int, unsigned int>& to ) const
    {
      if ( root ) {
        char tmp_key[ LOTreeNode::one_key_len ];
        root->getAllFeatureCollins( trie, to, tmp_key );
      }
    }
    */

    typedef 
    std::map< std::vector< std::pair< int, int> >, int >  
    production_map_t;

    void makeCollinsPProductionMap( production_map_t& to ) const {
      if ( root ) {
        root->makeCollinsPProductionMap( to );
      }
    }

    LOTree* makeCollinsPLOTree( production_map_t& prod_map ) const
    {
      if ( root ) {
        LOTreeNode* col_root = root->makeCollinsPLOTree( prod_map );
        if ( col_root ) {
          return new LOTree( col_root );
        }
      }
      return 0;
    }
    
  };

};

#endif
