/*
 TinySVM -- Yet Another Tiny SVM Package

 $Id: model.cpp,v 1.4 2002/08/20 06:31:17 taku-ku Exp $;

 Copyright (C) 2001-2002  Taku Kudo <taku-ku@is.aist-nara.ac.jp>
 All rights reserved.

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Library General Public
 License as published by the Free Software Foundation; either
 version 2 of the License, or (at your option) any later verjsion.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Library General Public License for more details.

 You should have received a copy of the GNU Library General Public
 License along with this library; if not, write to the
 Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 Boston, MA 02111-1307, USA.
*/


#include "model.h"
#include "common.h"
#include "kernel.h"
#include "classifier.h"
#include "qp_solver.h"
#include "example.h"

#include <fstream>

// $Id: model.cpp,v 1.4 2002/08/20 06:31:17 taku-ku Exp $;

#define MAKE_KERNEL { if (!kernel) \
      kernel = new Classifier( *(dynamic_cast <BaseExample *>(this)), param); };

namespace TinySVM {

Model::Model ()
{
  margin = vc = sphere = 0;
  svindex_size = bsv = -1;
  kernel = 0;
  alpha = 0;
  G = 0;
  feature_type = BINARY_FEATURE;
  verbose = 0; // J.K
}

Model::Model (const Param & p)
{
  margin = vc = sphere = 0;
  svindex_size = bsv = -1;
  kernel = 0;
  alpha = 0;
  G = 0;
  feature_type = BINARY_FEATURE;
  param = p;
  verbose = 0; // J.K
  A = 1.0;
  B = 1.0;
}


Model::~Model ()
{
  delete kernel;
}

int
Model::clear ()
{
  delete kernel;
  kernel = 0;
  alpha  = 0;
  G      = 0;
  return BaseExample::clear ();      
  A = 1.0;
  B = 1.0;
}

Model & Model::operator = (Model &m)
{
  if (this != &m) {
    BaseExample::operator = (m);
    margin = vc = 0;
    feature_type = BINARY_FEATURE;
    param = m.param;
    b     = m.b;
    bsv   = m.bsv;
    delete kernel;
    kernel = 0;
    verbose = m.verbose; // J.K
    A = m.A;
    B = m.B;
  }

  return *this;
}

double
Model::classify (const char *s, bool prob)
{
  MAKE_KERNEL;
  if ( !prob ) {
    return kernel->getDistance (s) - b;
  } else {
    double dist = kernel->getDistance (s) - b;
    return 1.0/(1.0+exp( A*dist + B ));
  }
}

double
Model::classify (const feature_node * f, bool prob)
{
  MAKE_KERNEL;
  if ( !prob ) {
    return kernel->getDistance (f) - b;
  } else {
    double dist = kernel->getDistance (f) - b;
    return 1.0/(1.0+exp( A*dist + B ));
  }
}

double
Model::estimateMargin ()
{
  if (margin) return margin;

  MAKE_KERNEL;
  double m = 0;
  for (int i = 0; i < l; i++) 
    m += y[i] * kernel->getDistance (x[i]);

  return (margin = 1.0 / sqrt (m));
}

double
Model::estimateVC ()
{
  if (vc) return vc;
  if (!margin) margin = estimateMargin ();
  if (!sphere) sphere = estimateSphere();
  return (vc = (sphere * sphere) / (margin * margin) + 1);
}

double
Model::estimateXA (const double rho)
{
  MAKE_KERNEL;

 // calculate R_delta, Roughly axpproxmimated
 feature_node *org = new feature_node [1];
 org[0].index = -1;
 org[0].value = 0;

 // I think it is good axproximation for binary vector
 double rsq = -INF;
 for (int i = 0; i < l; i++) {
   rsq = _max(rsq, kernel->getKernel (x[i], x[i]) - kernel->getKernel (x[i], org));
 }

 double result = 0.0;
 for (int i = 0; i < l; i++) {
   double d = kernel->getDistance (x[i]) - b;
   double eps;
   double alpha;
   if (y[i] > 0) { eps = _max(0.0, 1 - d); alpha =  y[i]; }
   else          { eps = _max(0.0, 1 + d); alpha = -y[i]; };
   if ((rho * alpha * rsq + eps) >= 1.0) result++;
 }

 delete [] org;

 return result;
}

double 
Model::estimateSphere()
{
  MAKE_KERNEL;

  feature_node *org = new feature_node[1];
  org[0].index = -1;
  org[0].value = 0;
  double orgsq = kernel->getKernel (org, org);

  double maxd = -INF;
  for (int i = 0; i < l; i++) {
    maxd =
      _max (maxd,kernel->getKernel (x[i],x[i]) - 2 * kernel->getKernel (x[i], org) + orgsq);
  }

  delete [] org;

  return (sphere = sqrt(maxd));
}

int
Model::compress ()
{
  if (param.degree != 1) {
    fprintf(stderr, "Model::compress: cannot compress model.\n");
    return 0;
  }

  double *w = new double [d + 1];
  for (int i = 0; i <= d; i++) w[i] = 0.0;
  for (int i = 0; i < l; i++) {
    for (feature_node *node = x[i]; node->index >= 0; node++) 
      w[node->index] += param.param_s * y[i] * node->value;
  }

  pack_d = 0;
  for (int i = 0; i <= d; i++) if (w[i] != 0) pack_d++;
  feature_node *fn = new feature_node [pack_d + 1];

  int j = 0;
  for (int i = 0; i <= d; i++) {
    if (w[i] != 0) {
      fn[j].index = i;
      fn[j].value = w[i];
      j++;
    }
  }
  fn[j].index = -1;
  fn[j].value = 0;

  delete [] w;
  clear();

  return add(1, fn);
}

int
Model::read (const char *filename, const char *mode, const int offset)
{
  FILE *fp = fopen (filename, mode);
  if (!fp) return 0;

  fseek (fp, offset, SEEK_SET);
  char *buf;
  char tmpbuf[1024];
  char version[512];
  int tmpl;
  // read parameter
  fscanf (fp, "%s Version %s%*[^\n]\n", tmpbuf, version);
  fscanf (fp, "%d%*[^\n]\n",  &param.kernel_type);
  fscanf (fp, "%d%*[^\n]\n",  &param.degree);
  fscanf (fp, "%lf%*[^\n]\n", &param.param_g);
  fscanf (fp, "%lf%*[^\n]\n", &param.param_s);
  fscanf (fp, "%lf%*[^\n]\n", &param.param_r);
  fscanf (fp, "%d%*[^\n]\n",  &param.geometric_type); // J.K
  fscanf (fp, "%d%*[^\n]\n",  &param.node_restriction_type); // J.K
  fscanf (fp, "%lf%*[^\n]\n", &param.restriction_ratio); // J.K
  fscanf (fp, "%lf%*[^\n]\n", &param.depth_factor); // J.K
  fscanf (fp, "%d%*[^\n]\n",  &param.self_normalize); // J.K
  fscanf (fp, "%d%*[^\n]\n",  &param.threshold_normalize); // J.K
  fscanf (fp, "%lf%*[^\n]\n", &param.threshold_normalize_threshold); // J.K
  fscanf (fp, "%lf%*[^\n]\n", &A); // J.K
  fscanf (fp, "%lf%*[^\n]\n", &B); // J.K
  fscanf (fp, "%d%*[^\n]\n",  &param.ignore_malicious); // J.K
  fscanf (fp, "%s%*[^\n]\n",  tmpbuf);

  std::cerr << "param.ignore_malicious = " << param.ignore_malicious << "\n" << std::flush;
  std::cerr << "param.itrie            = " << param.itrie << "\n" << std::flush;
  
  param.dot_kernel = 1;
  if (param.kernel_type == RBF || param.kernel_type == ANOVA || Param::is_tree_kernel( param.kernel_type) ) param.dot_kernel = 0; // J.K.

  if ( Param::is_tree_kernel( param.kernel_type) ) {
    std::cerr << "learning with a tree kernel\n" << std::flush;
  }
  if ( Param::is_node_lookup_kernel( param.kernel_type) ) {
    std::cerr << "learning with a fast node-lookup kernel\n" << std::flush;
  }

  buf = readLine(fp);
  if (sscanf (buf, "%d %d %d %lf%%*[^\n]\n", &tmpl, &bsv, &svindex_size, &loss) != 4) {
    sscanf (buf, "%d%*[^\n]\n", &tmpl); // old version
    svindex_size = -1;
    loss = -1;
  }

  tmpl--;
  fscanf (fp, "%lf%*[^\n]\n", &b);

  int line = 0;
  while ((buf = readLine (fp)) != NULL && line < tmpl) {
    line++;
    if (!this->add (buf)) {
      fprintf (stderr, "WARNING: Format Error in file [%s] line [%d]\n",
	       filename, line);
      fclose (fp);
      return 0;
    }
  }
  fclose (fp);

  param.verbose = verbose;

  // Read tree feature weights if they are exist
  {
    std::string fw_name( filename );
    fw_name += ".fw";
    std::ifstream ifs( fw_name.c_str() );
    readFeatureWeightFromStream( ifs );
    ifs.close();
  }
  // Read production map for Collins' kernel if it exist
  {
    std::string collprod_name( filename );
    collprod_name += ".collprod";
    std::ifstream ifs( collprod_name.c_str() );
    if ( ifs ) {
      readCollinsProductionMapFromStream( ifs );
      ifs.close();
    }
  }

  // Read tree feature trie if it exists
  {
    std::string ft_name( filename );
    ft_name += ".trie";
    std::ifstream ifs( ft_name.c_str() );
    if ( ifs ) {
      ifs.close();
      std:: cerr << "Read feature trie ...\n" << flush;
      feature_trie = new trie_t();
      if ( feature_trie->open( ft_name.c_str() ) < 0 ){
        std:: cerr << "Read feature trie failed.\n" << flush;
        exit(1);
      }
      std:: cerr << " ... end. (size = " << feature_trie->getSize() << ")\n" << flush;
      
      // check 
      if ( false ) {
        for ( int i = 0; i < l; i++ ) {
          feature_node* node = x[i];
          if ( node->extra && node->extra->getTreeVec() ) {
            feature_node* cor_fv = node->extra->getTreeVec();
            std::map< unsigned int, unsigned int > fmap;
            if ( param.kernel_type == TREE_COLLINS ||
                 param.kernel_type == TREE_COLLINS_FAST ) {
              //node->extra->getTree()->getAllFeatureCollins( feature_trie, fmap );
            } else {
              //node->extra->getTree()->getAllFeature( feature_trie, fmap );
            }
            feature_node* fv = new feature_node[ fmap.size() + 1 ];
            
            bool conv_fail = false;
            
            int cor_size = 0;
            for ( feature_node* n = cor_fv; n->index != -1; n++ ) {
              cor_size++;
            }
            if ( cor_size != fmap.size() ) {
              std::cerr << "cor_size = " << cor_size << " sys_size = " << fmap.size() << "\n" << std::flush;
              conv_fail = true;
            }
            int index = 0;
            for ( std::map< unsigned int, unsigned int >::const_iterator itr 
                    = fmap.begin(); itr != fmap.end(); itr++ ) {
              if ( index >= cor_size || index >= fmap.size() ) {
                break;
              }
              fv[ index ].index = itr->first;
              fv[ index ].value = itr->second;
              std::cerr << "cor = " << cor_fv[index].index  
                        << ":" << cor_fv[index].value;
              std::cerr << " sys = " << fv[index].index  << ":" << fv[index].value << "\n" << std::flush;
              
              if ( cor_fv[index].index != fv[index].index || 
                   cor_fv[index].value != fv[index].value ) {
                
                conv_fail = true;
              }
              index++;
            }
            if ( conv_fail ) {
              std::cerr << "conversion failed.\n" << std::flush;
              exit(1);
            }
            fv[ index ].index = -1;
            fv[ index ].value = 0.0;
            
            delete[] fv;
          }
        }
      }
    }
  }

  // make kernel
  MAKE_KERNEL;

  return 1;
}

int
Model::write (const char *filename, const char *mode, const int offset)
{
  FILE *fp = fopen (filename, mode);
  if (!fp) return 0;

  //   fprintf(fp,"SVM-light Version V3.02\n");
  fprintf (fp, "TinySVM Version %s\n", VERSION);
  fprintf (fp, "%d # kernel type\n", param.kernel_type);
  fprintf (fp, "%d # kernel parameter -d\n", param.degree);
  fprintf (fp, "%.8g # kernel parameter -g\n", param.param_g);
  fprintf (fp, "%.8g # kernel parameter -s\n", param.param_s);
  fprintf (fp, "%.8g # kernel parameter -r\n", param.param_r);
  // J.K
  fprintf (fp, "%d # geometric type\n", param.geometric_type);
  fprintf (fp, "%d # node restriction type\n", param.node_restriction_type);
  fprintf (fp, "%.8g # restriction ratio\n", param.restriction_ratio );
  fprintf (fp, "%.8g # depth factor\n", param.depth_factor );
  fprintf (fp, "%d # do self normalize\n", (int)param.self_normalize );
  fprintf (fp, "%d # do threshold normalize\n", (int)param.threshold_normalize );
  fprintf (fp, "%.8g # threshold normalize threshold\n", param.threshold_normalize_threshold );
  fprintf (fp, "%.8g # A of probfit\n", A );
  fprintf (fp, "%.8g # B of probfit\n", B );
  fprintf (fp, "%d # ignore malicious\n", (int)param.ignore_malicious );
  fprintf (fp, "empty # kernel parameter -u\n");

  // add L1loss, # of training size to the model file since version 0.02
  fprintf (fp, "%d %d %d %.8g # number of SVs/BSVs/number of training data/L1 loss \n", 
	   l+1, bsv, svindex_size, loss);   // triky, svm_light must be l+1

  fprintf (fp, "%.16g # threshold b\n", b);

  for (int i = 0; i < l; i++) {
    /*
    fprintf (fp, "%.16g", y[i]);
    for (feature_node * node = x[i]; node->index >= 0; node++)
      fprintf (fp, " %d:%.16g", node->index, node->value);
    fprintf (fp, "\n");
    */
    fprintf (fp, "%.16g", y[i]);
    for (feature_node * node = x[i]; node->index >= 0; node++) {
      fprintf (fp, " %d:%.16g", node->index, node->value);
    }
    if ( x[i]->extra ) { // J.K.
      fprintf (fp, " %s", x[i]->extra->toString().c_str());
    }
    fprintf (fp, "\n");
  }

  fclose (fp);
  return 1;
}
}
