/*
 TinySVM -- Yet Another Tiny SVM Package

 $Id: svm_classify.cpp,v 1.4 2002/08/20 06:31:17 taku-ku Exp $;

 Copyright (C) 2001-2002  Taku Kudo <taku-ku@is.aist-nara.ac.jp>
 All rights reserved.

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Library General Public
 License as published by the Free Software Foundation; either
 version 2 of the License, or (at your option) any later verjsion.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Library General Public License for more details.

 You should have received a copy of the GNU Library General Public
 License along with this library; if not, write to the
 Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 Boston, MA 02111-1307, USA.
*/


// modified for J.K.'s use

#include "common.h"
#include "misc.h"
#include "model.h"
#include "example.h"
#include "base_example.h"
#include "kernel.h"
#include "param.h"

const char *help_message =
"\nUsage: svm_classify [options] test-file model-file out-file\n\n\
options:\n\
  -V, --verbose          set verbose mode.\n\
  -v, --version          show the version of TinySVM and exit.\n\
  -h, --help             show this help and exit.\n\
  -N, --num              maximum number of sample to classify\n\
  -P, --prob             probability output\n\
  -I, --ignore-trie      ignore feature trie\n";

static const char *short_options = "VvhN:PI";

static struct option long_options[] = {
  {"verbose",     no_argument,       NULL, 'V'},
  {"version",     no_argument,       NULL, 'v'},
  {"help",        no_argument,       NULL, 'h'},
  {"num",         required_argument, NULL, 'N'},
  {"prob",        no_argument,       NULL, 'P'},
  {"ignore-trie", no_argument,       NULL, 'I'},
  {NULL, 0, NULL, 0}
};

int
main (int argc, char **argv)
{
  FILE *fp;
  FILE *ofp;

  int verbose = 0;
  int num = -1;

  bool   prob   = false;
  bool   ignore_trie = false;

  while (1) {
    int opt = getopt_long (argc, argv, short_options, long_options, NULL);
    if (opt == EOF) break;

    switch (opt) {
    case 'V':
      verbose = 1;
      break;
    case 'v':
      fprintf (stdout, "%s of %s\n%s\n", argv[0], VERSION, COPYRIGHT);
      exit (EXIT_SUCCESS);
    case 'h':
      fprintf (stdout, "%s%s", COPYRIGHT, help_message);
      exit (EXIT_SUCCESS);
    case 'N':
      num = atoi( optarg );
      break;
    case 'P':
      prob = true;
      break;
    case 'I':
      ignore_trie = true;
      break;
    default:
      fprintf (stdout, "%s\n", COPYRIGHT);
      fprintf (stdout, "Try `svm_classify --help' for more information.\n");
      exit (EXIT_SUCCESS);
    }
  }

  if (argc < 4) {
    fprintf (stdout, "%s\n", COPYRIGHT);
    fprintf (stdout, "Try `svm_classify --help' for more information.\n");
    exit (EXIT_FAILURE);
  }

  TinySVM::Model model;
  model.verbose = verbose;
  model.param.itrie = ignore_trie;
  std::cerr << "A=" << ignore_trie << "\n" << std::flush;
  std::cerr << "B=" << model.param.itrie << "\n" << std::flush;
  if (!model.read (argv[argc - 2])) {
    fprintf (stderr, "%s: %s: No such file or directory\n", argv[0], argv[argc - 2]);
    exit (EXIT_FAILURE);
  }

  if ( (!strcmp (argv[argc - 3], "-")) || (!strcmp (argv[argc - 3], "stdin")) ) {
    fp = stdin;
  } else {
    if ((fp = fopen (argv[argc - 3], "r")) == NULL) {
      fprintf (stderr, "%s: %s: No such file or directory\n", argv[0], argv[argc - 3]);
      exit (EXIT_FAILURE);
    }
  }

  if ( (!strcmp (argv[argc - 1], "-")) || (! strcmp(argv[argc - 1], "stdout")) ) {
    ofp = stdout;
  } else {
    if ((ofp = fopen (argv[argc - 1], "w")) == NULL) {
      fprintf (stderr, "%s: %s: No such file or directory\n", argv[0], argv[argc - 1]);
      exit (EXIT_FAILURE);
    }
  }
  

  char *line;
  int all = 0;
  int correct = 0;
  int res_a = 0;
  int res_b = 0;
  int res_c = 0;
  int res_d = 0;

  while ( true ) {
    if ( num >= 0 && num <= all ) {
      break;
    }
    line = model.readLine (fp);
    if (line == NULL) {
      break;
    }

    if (verbose ) {
      std::cerr << "line=" << line << "\n" << std::flush;
    }
    double y = 0;
    int len = strlen (line);
    int i = 0;
    while (isspace (line[i])) i++;
    y = atof (line + i);
    while (i < len && !isspace (line[i])) i++;
    while (i < len && isspace (line[i])) i++;
    double out_val 
      = model.classify ((const char *) (line + i), prob );

    double pred    = 0.0;
    if ( prob ) {
      if ( out_val > 0.5 ) {
        pred = 1.0;
      } else {
        pred = -1.0;
      }
    } else {
      if ( out_val > 0.0 ) {
        pred = 1.0;
      } else {
        pred = -1.0;
      }
    }

    if (verbose) printf ("%g %g\n", y, out_val);
    fprintf(ofp, "%f\n", out_val);
    fflush (ofp);

    all++;
    if(pred > 0) {
      if(y > 0) correct++;
      if(y > 0) res_a++; else res_b++;
    } else {
      if(y < 0) correct++;
      if(y > 0) res_c++; else res_d++;
    }
    
    
  }
  fprintf (stderr, "Accuracy:   %.5f%% (%d/%d)\n", 100.0 * correct/all , correct, all);
  fprintf (stderr, "Precision:  %.5f%% (%d/%d)\n", 100.0 * res_a/(res_a + res_b), res_a, res_a + res_b);
  fprintf (stderr, "Recall:     %.5f%% (%d/%d)\n", 100.0 * res_a/(res_a + res_c), res_a, res_a + res_c);
  fprintf (stderr, "System/Answer p/p p/n n/p n/n: %d %d %d %d\n", res_a,res_b,res_c,res_d);
  
  fclose (fp);
 
  if ( ofp != stdout ) {
    fclose( ofp );
  }

  return EXIT_SUCCESS;
}
