/*
 This file is added by J. Kazama to the TinySVM package.
 Copyright (C) 2005. Jun'ichi Kazama <kazama@jaist.ac.jp>
 All rights reserved.

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Library General Public
 License as published by the Free Software Foundation; either
 version 2 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Library General Public License for more details.

 You should have received a copy of the GNU Library General Public
 License along with this library; if not, write to the
 Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 Boston, MA 02111-1307, USA.
*/

#include "freqtm.h"
#include "misc.h"
#include "common.h"
#include "example.h"

#include <string>
#include <fstream>

using namespace std;
using namespace TinySVM;

const char *help_message =
"\nUsage: tree2fv [options] example-file out-file\n\n\
options:\n\
  -V, --verbose          set verbose mode.\n\
  -v, --version          show the version of TinySVM and exit.\n\
  -h, --help             show this help and exit.\n\
  -t, --kernel-type      select type of tree kernel function.\n\
                         (see the help of svm_learn)\n\
  -b, --restriction=INT  node restriction type of the tree kernel.\n\
                         (see the help of svm_learn)\n\
  -u, --depth-factor     depth factor of tree kernel.\n\
  -m, --minsup           minimum support of FREQT(M).\n\
                         (set -m 2 for the exact FREQTM)\n\
  -M, --minpat           minimum pattern size (in \# of nodes).\n\
                         (usually you don't need to set this.)\n\
  -L, --maxpat           maximum pattern size (in \# of nodes).\n\
  -D, --malicious-size   the size of subtrees considered malicious.\n\
  -R, --malicious-rate   if |occ(F_k)/sup(F_i)| > 1/R then stop searching.\n\
                         (note that this is the inverse of R in my HLT-EMNLP paper.)\n\
  -d, --heavy-check      do pruning (P2) in the HLT-EMNLP paper (heavy check)\n\
  -p, --pre-check        do pruning (PC) in the HLT-EMNLP paper\n\
                         (will double the conversion time in exchange of a little bit shorter feature vectors.)\n\
  -k, --mark-pruning     do pruning exploiting the mark information in the tree.\n\
                         (have effect only when the kernel is for marked trees, i.e., -b 1.)\n\
  -c, --compress         do feature compression in the HLT-EMNLP paper.\n\
  -T, --track-pattern    track feature subtree patterns.\n\
                         (needed for -O)\n\
  -O, --output-pattern   output the feature subtree patterns.\n\
                         (the output can be used, e.g., to construct a feature subtree trie.)\n\
\n\
  SO, the typical usage as FREQTM is:\n\
  > tree2fv -m 2 -D 20 -R 0.05 -t 9 -b 1 -u 0.2 -p -d -c A2.ex A2.ex.fv\n\
  THEN, the training using A2.ex.fv instead of A2.ex is sped up automatically.\n\
  > svm_learn -t 9 -b 1 -u 0.2 A2.ex (SLOW; using DP for all pairs)\n\
  > svm_learn -t 9 -b 1 -u 0.2 A2.ex.fv (FAST; using DP only for malicious pairs)\n";

static const char *short_options = "Vvht:b:u:m:M:L:D:R:dpcTkO";

static struct option long_options[] = {
  {"verbose",   no_argument,       NULL, 'V'},
  {"version",   no_argument,       NULL, 'v'},
  {"help",      no_argument,       NULL, 'h'},
  {"kernel-type",  required_argument, NULL, 't'},
  {"restriction",  required_argument, NULL, 'b'},
  {"depth-factor", required_argument, NULL, 'u'},
  {"minsup", required_argument, NULL, 'm'},
  {"minpat", required_argument, NULL, 'M'},
  {"maxpat", required_argument, NULL, 'L'},
  {"malicious-size", required_argument, NULL, 'D'},
  {"malicious-rate", required_argument, NULL, 'R'},
  {"heavy-check", no_argument, NULL, 'd'},
  {"pre-check",   no_argument, NULL, 'p'},
  {"compress",    no_argument, NULL, 'c'},
  {"track-pattern", no_argument, NULL, 'T'},
  {"output-pattern", no_argument, NULL, 'O'},
  {"mark-pruning", no_argument, NULL, 'k'},
  {NULL, 0, NULL, 0}
};

int
main (int argc, char **argv)
{
  extern char *optarg;
  int verbose = 0;
  unsigned int minsup = 1;
  unsigned int minpat = 1;
  unsigned int maxpat         = 0xffffffff;
  unsigned int malicious_size = 0xffffffff;
  double       malicious_rate = 0.0;

  bool do_precheck         = false;
  bool do_heavy_check      = false;
  bool do_feature_compress = false;
  bool do_track_pattern    = false;
  bool do_output_pattern   = false;
  bool do_mark_pruning     = false;
  int  kernel_type         = 0;
  int  restriction_type    = 0;
  vector<double> depth_factors;
  //double depth_factor      = 1.0;
  
  string input_fname;
  string output_fname;

  while (1) {
    int opt = getopt_long (argc, argv, short_options, long_options, NULL);
    if (opt == EOF) break;

    switch (opt) {
    case 'V':
      verbose = 1;
      break;
    case 'v':
      fprintf (stdout, "%s of %s\n", argv[0], VERSION);
      exit (EXIT_SUCCESS);
    case 'h':
      fprintf (stdout, "%s", help_message);
      exit (EXIT_SUCCESS);
    case 't':
      kernel_type = atoi (optarg);
      break;
    case 'b':
      restriction_type = atoi ( optarg );
      break;
    case 'u':
      {
        string s( optarg );
        string::iterator i = s.begin();
        if ( s.find( ':' ) != string::npos ) { // like 0.1:0.2:0.3
          while ( i < s.end() ) {
            string::iterator j = find( i, s.end(), ':' );
            string us( i, j );
            if ( us != "" ) {
              double d = atof( us.c_str() );
              cerr << "u=" << d << " ";
              depth_factors.push_back( d );
            }
            i = j + 1;
          }
          cerr << "\n" << flush;
        } else {
          depth_factors.push_back( atof( optarg ) );
        }
      }
      break;
    case 'm':
      minsup = atoi (optarg);
      break;
    case 'M':
      minpat = atoi (optarg);
      break;
    case 'L':
      maxpat = atoi (optarg);
      break;
    case 'D':
      malicious_size = atoi( optarg );
      break;
    case 'R':
      malicious_rate = atof( optarg );
      break;
    case 'd':
      do_heavy_check = true;
      break;
    case 'p':
      do_precheck    = true;
      break;
    case 'c':
      do_feature_compress = true;
      break;
    case 'T':
      do_track_pattern = true;
      break;
    case 'O':
      do_output_pattern = true;
      break;
    case 'k':
      do_mark_pruning = true;
      break;
    default:
      fprintf (stdout, "Try `tree2fv --help' for more information.\n");
      exit (EXIT_SUCCESS);
    }
  }
  cerr << "argc = " << argc << "\n" << flush;
  if (argc < 3 ) {
    fprintf (stdout, "Try `tree2fv --help' for more information.\n");
    exit (EXIT_FAILURE);
  }

  input_fname   = argv[ argc - 2];
  output_fname  = argv[ argc - 1];

  if ( depth_factors.size() == 0 ) {
    depth_factors.push_back( 0.0 );
  }
#ifdef LABEL_INTEGER
  cerr << "(Integer Label)\n" << flush;
#else
  cerr << "(String Label)\n"  << flush;
#endif
  cerr << "input:  " << input_fname   << "\n" << flush;
  cerr << "output: " << output_fname  << "\n" << flush;
  cerr << "minsup: " << minsup        << "\n" << flush;
  cerr << "minpat: " << minpat        << "\n" << flush;
  cerr << "maxpat: " << maxpat        << "\n" << flush;
  cerr << "malicious_size:      " << malicious_size << "\n" << flush;
  cerr << "malicious_rate:      " << malicious_rate << "\n" << flush;
  cerr << "do_precheck:         " << do_precheck    << "\n" << flush;
  cerr << "do_heavy_check:      " << do_heavy_check << "\n" << flush;
  cerr << "do_feature_compress: " << do_feature_compress << "\n" << flush;
  cerr << "do_track_pattern:    " << do_track_pattern    << "\n" << flush;
  cerr << "do_output_pattern:   " << do_output_pattern   << "\n" << flush;
  cerr << "do_mark_pruning:     " << do_mark_pruning     << "\n" << flush;
  cerr << "kernel_type:         " << kernel_type         << "\n" << flush;
  cerr << "restriction_type:    " << restriction_type    << "\n" << flush;
  //cerr << "depth_factor:        " << depth_factor        << "\n" << flush;
  cerr << "depth_factors:       ";
  for ( int i = 0; i < depth_factors.size(); i++ ) {
    cerr << depth_factors[i] << " ";
  }
  cerr << "\n" << flush;

  double depth_factor = depth_factors[0];

  Example example;
  if (!example.read ( input_fname.c_str() )) {
    fprintf (stderr, "%s: %s: No such file or directory\n", argv[0],
	     input_fname.c_str() );
    exit (EXIT_FAILURE);
  }

  FreqtmBase* freqtm = 0;

  typedef Freqtm< false, false >  FreqtmInst;   // Kashima, w/o mark
  typedef Freqtm< true,  false >  FreqtmInstM;  // Kashima, w/  mark
  typedef Freqtm< true,  true  >  FreqtmInstMP; // Kashima, w/ mark + markpruning

  typedef FreqtmCollins< false, false > FreqtmCollinsInst;
  typedef FreqtmCollins< true, false >  FreqtmCollinsInstM;
  typedef FreqtmCollins< true, true  >  FreqtmCollinsInstMP;

  if (kernel_type == TREE || kernel_type == TREE_FAST ) {
    if ( restriction_type == NO_RESTRICT ) {
      freqtm 
        = new FreqtmInst(example, minsup, minpat, maxpat, 
                         malicious_size,
                         malicious_rate,
                         do_precheck, do_heavy_check, 
                         do_feature_compress,
                         do_track_pattern,
                         do_output_pattern,
                         depth_factor, 1);
    } else if ( restriction_type == RESTRICT_1 ) {
      if ( !do_mark_pruning ) {
        freqtm = new FreqtmInstM(example, minsup, minpat, maxpat, 
                                 malicious_size,
                                 malicious_rate,
                                 do_precheck, do_heavy_check, 
                                 do_feature_compress,
                                 do_track_pattern,
                                 do_output_pattern,
                                 depth_factor, 1);
      } else if ( do_mark_pruning ) {
        freqtm = new FreqtmInstMP(example, minsup, minpat, maxpat, 
                                  malicious_size,
                                  malicious_rate,
                                  do_precheck, do_heavy_check, 
                                  do_feature_compress,
                                  do_track_pattern,
                                  do_output_pattern,
                                  depth_factor, 1);
      }
    }
  } else if (kernel_type    == TREE_COLLINS 
             || kernel_type == TREE_COLLINS_FAST ) {
    if ( restriction_type == NO_RESTRICT ) {

      freqtm = new FreqtmCollinsInst(example, minsup, minpat, maxpat, 
                                     malicious_size,
                                     malicious_rate,
                                     do_precheck, do_heavy_check, 
                                     do_feature_compress,
                                     do_track_pattern, 
                                     do_output_pattern,
                                     depth_factor, 1 );
    } else if ( restriction_type == RESTRICT_1 ) {
      if ( !do_mark_pruning ) {
         freqtm = new FreqtmCollinsInstM(example, minsup, minpat, maxpat, 
                                         malicious_size,
                                         malicious_rate,
                                         do_precheck, do_heavy_check, 
                                         do_feature_compress,
                                         do_track_pattern, 
                                         do_output_pattern,
                                         depth_factor, 1 );
      } else if ( do_mark_pruning ) {
        freqtm = new FreqtmCollinsInstMP(example, minsup, minpat, maxpat, 
                                         malicious_size,
                                         malicious_rate,
                                         do_precheck, do_heavy_check, 
                                         do_feature_compress,
                                         do_track_pattern, 
                                         do_output_pattern,
                                         depth_factor, 1 );
      }
    }
  } else if ( ( kernel_type == TREE_COLLINS_P ||
                kernel_type == TREE_COLLINS_P_FAST ) 
              && restriction_type == RESTRICT_1 ) {
    
    cerr << "Unsupported Kernel & Restriction\n" << flush;
    exit(1);
  } else {
    cerr << "Unsupported Kernel & Restriction\n" << flush;
    exit(1);
  }

  ofstream* feature_weight_stream 
    = new ofstream( (output_fname + ".fw").c_str() );
  freqtm->setFeatureWeightStream( feature_weight_stream );

  ofstream* pattern_out_stream = 0;
  if ( do_output_pattern ) {
    pattern_out_stream = new ofstream( (output_fname + ".pat").c_str() );
    freqtm->setPatternOutStream( pattern_out_stream );
  }

  freqtm->run();

  example.write( output_fname.c_str() );

  feature_weight_stream->close();
  if ( pattern_out_stream ) {
    pattern_out_stream->close();
  }

  if ( example.production_map ) {
    ofstream ofs( (output_fname + ".collprod").c_str() );
    example.writeCollinsProductionMapToStream( ofs );
    ofs.close();
  }

  return 0;
}

